import io
import time
import gzip

import requests
import numpy as np
import pandas as pd

from linora import gfile
from linora.utils._progbar import Progbar
from linora.data._compress import decompress
from linora.data._utils import assert_dirs, get_file
from linora.image._image_io import save_image
from linora.image._image_util import array_to_image
from linora.data.Dataset._dataset import from_class_folder
from linora.data.datasets._config_path import Param

__all__ = ['mnist', 'mnist_fashion', 'mnist_kannada', 'mnist_tibetan',
           'mnist_kuzushiji10', 'mnist_kuzushiji49', 'mnist_kuzushiji_kanji']


def mnist(root=None, dataset=True, verbose=1):
    """MNIST handwritten digits dataset from http://yann.lecun.com/exdb/mnist
    
    Each sample is an gray image (in 3D NDArray) with shape (28, 28, 1).
    
    Attention: if exist dirs `root/mnist`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist data: 
    `root/mnist/train/0/xx.png`
    `root/mnist/train/2/xx.png`
    `root/mnist/train/6/xx.png`
    `root/mnist/test/0/xx.png`
    `root/mnist/test/2/xx.png`
    `root/mnist/test/6/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist`,
              root should be `/user/.../mydata`.
        dataset: whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist`.
    """
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist')
    p.add(1)
    for url in Param.mnist:
        get_file(url, gfile.path_join(task_path, url.split('/')[-1]), verbose=0)
        p.add(1)
    with gzip.open(gfile.path_join(task_path, 'train-labels-idx1-ubyte.gz'), 'rb') as lbpath:
        train_label = np.frombuffer(lbpath.read(), np.uint8, offset=8)

    with gzip.open(gfile.path_join(task_path, 'train-images-idx3-ubyte.gz'), 'rb') as imgpath:
        train = np.frombuffer(imgpath.read(), np.uint8, offset=16).reshape(len(train_label), 28, 28)

    with gzip.open(gfile.path_join(task_path, 't10k-labels-idx1-ubyte.gz'), 'rb') as lbpath:
        test_label = np.frombuffer(lbpath.read(), np.uint8, offset=8)

    with gzip.open(gfile.path_join(task_path, 't10k-images-idx3-ubyte.gz'), 'rb') as imgpath:
        test = np.frombuffer(imgpath.read(), np.uint8, offset=16).reshape(len(test_label), 28, 28)
    p.add(1)
    for i in set(train_label):
        gfile.makedirs(gfile.path_join(task_path, 'train', str(i)))
    for i in set(test_label):
        gfile.makedirs(gfile.path_join(task_path, 'test', str(i)))
    p.add(1)
    for idx in range(train.shape[0]):
        save_image(gfile.path_join(task_path, 'train', str(train_label[idx]), str(idx)+'.png'), 
                   array_to_image(train[idx].reshape(28, 28, 1)))
    p.add(1)
    for idx in range(test.shape[0]):
        save_image(gfile.path_join(task_path, 'test', str(test_label[idx]), str(idx)+'.png'), 
                   array_to_image(test[idx].reshape(28, 28, 1)))
    p.add(1)
    for url in Param.mnist:
        gfile.remove(gfile.path_join(task_path, url.split('/')[-1]))
    p.add(1)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
                .join({'test':from_class_folder(gfile.path_join(task_path, 'test'), label_encoder=1)}))
    return task_path


def mnist_fashion(root=None, dataset=True, verbose=1):
    """A dataset of Zalando's article images consisting of fashion products.
    
    Fashion mnist datasets is a drop-in replacement of the original MNIST dataset
    from https://github.com/zalandoresearch/fashion-mnist.
    Each sample is an gray image (in 3D NDArray) with shape (28, 28, 1).
    
    Attention: if exist dirs `root/mnist_fashion`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist_fashion data: 
    `root/mnist_fashion/train/0/xx.png`
    `root/mnist_fashion/train/2/xx.png`
    `root/mnist_fashion/train/6/xx.png`
    `root/mnist_fashion/test/0/xx.png`
    `root/mnist_fashion/test/2/xx.png`
    `root/mnist_fashion/test/6/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist_fashion`,
              root should be `/user/.../mydata`.
        dataset: whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist_fashion`.
    """
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist_fashion')
    p.add(1)
    for url in Param.mnist_fashion:
        get_file(url, gfile.path_join(task_path, url.split('/')[-1]), verbose=0)
        p.add(1)
    with gzip.open(gfile.path_join(task_path, 'train-labels-idx1-ubyte.gz'), 'rb') as lbpath:
        train_label = np.frombuffer(lbpath.read(), np.uint8, offset=8)

    with gzip.open(gfile.path_join(task_path, 'train-images-idx3-ubyte.gz'), 'rb') as imgpath:
        train = np.frombuffer(imgpath.read(), np.uint8, offset=16).reshape(len(train_label), 28, 28)

    with gzip.open(gfile.path_join(task_path, 't10k-labels-idx1-ubyte.gz'), 'rb') as lbpath:
        test_label = np.frombuffer(lbpath.read(), np.uint8, offset=8)

    with gzip.open(gfile.path_join(task_path, 't10k-images-idx3-ubyte.gz'), 'rb') as imgpath:
        test = np.frombuffer(imgpath.read(), np.uint8, offset=16).reshape(len(test_label), 28, 28)
    p.add(1)
    for i in set(train_label):
        gfile.makedirs(gfile.path_join(task_path, 'train', str(i)))
    for i in set(test_label):
        gfile.makedirs(gfile.path_join(task_path, 'test', str(i)))
    p.add(1)
    for idx in range(train.shape[0]):
        save_image(gfile.path_join(task_path, 'train', str(train_label[idx]), str(idx)+'.png'), 
                   array_to_image(train[idx].reshape(28, 28, 1)))
    p.add(1)
    for idx in range(test.shape[0]):
        save_image(gfile.path_join(task_path, 'test', str(test_label[idx]), str(idx)+'.png'), 
                   array_to_image(test[idx].reshape(28, 28, 1)))
    p.add(1)
    for url in Param.mnist_fashion:
        gfile.remove(gfile.path_join(task_path, url.split('/')[-1]))
    p.add(1)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
                .join({'test':from_class_folder(gfile.path_join(task_path, 'test'), label_encoder=1)}))
    return task_path


def mnist_kannada(root=None, dataset=True, verbose=1):
    """kannada-MNIST from https://github.com/vinayprabhu/Kannada_MNIST.
    
    The Kannada-MNIST dataset was created an a drop-in substitute for the standard MNIST dataset.
    
    Each sample is an gray image (in 3D NDArray) with shape (28, 28, 1).
    
    Attention: if exist dirs `root/mnist_kannada`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist_kannada data: 
    `root/mnist_kannada/train/0/xx.png`
    `root/mnist_kannada/train/2/xx.png`
    `root/mnist_kannada/train/6/xx.png`
    `root/mnist_kannada/test/0/xx.png`
    `root/mnist_kannada/test/2/xx.png`
    `root/mnist_kannada/test/6/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist_kannada`,
              root should be `/user/.../mydata`.
        dataset: bool, whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist_kannada` or la.data.Dataset.
    """
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist_kannada')
    p.add(1)
    zip_path = get_file(Param.mnist_kannada[0], task_path+'/kannada_MNIST.zip', verbose=0)
    p.add(2)
    unzip_path = decompress(task_path+'/kannada_MNIST.zip')
    p.add(1)
    train = pd.read_csv(gfile.path_join(task_path, 'kannada_MNIST/kannada_MNIST_train.csv'), header=None, dtype='uint8')
    test = pd.read_csv(gfile.path_join(task_path, 'kannada_MNIST/kannada_MNIST_test.csv'), header=None, dtype='uint8')
    p.add(2)
    for i in set(train[0]):
        gfile.makedirs(gfile.path_join(task_path, 'train', str(i)))
        gfile.makedirs(gfile.path_join(task_path, 'test', str(i)))
    p.add(1)
    for i in range(len(train)):
        save_image(gfile.path_join(task_path, 'train', str(train.iat[i, 0]), str(i)+'.png'),
                   array_to_image(train.iloc[i, 1:].values.reshape(28, 28, 1)))
    p.add(1)
    for i in range(len(test)):
        save_image(gfile.path_join(task_path, 'test', str(test.iat[i, 0]), str(i)+'.png'),
                       array_to_image(test.iloc[i, 1:].values.reshape(28, 28, 1)))
    p.add(1)
    gfile.remove(zip_path)
    gfile.remove(unzip_path)
    p.add(1)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
                .join({'test':from_class_folder(gfile.path_join(task_path, 'test'), label_encoder=1)}))
    return task_path


def mnist_kuzushiji10(root=None, dataset=True, verbose=1):
    """Kuzushiji-MNIST from https://github.com/rois-codh/kmnist.
    
    Kuzushiji-MNIST is a drop-in replacement for the
    MNIST dataset (28x28 grayscale, 70,000 images), 
    provided in the original MNIST format as well as a NumPy format.
    Since MNIST restricts us to 10 classes, we chose one character to
    represent each of the 10 rows of Hiragana when creating Kuzushiji-MNIST.
    
    Each sample is an gray image (in 3D NDArray) with shape (28, 28, 1).
    
    Attention: if exist dirs `root/mnist_kuzushiji10`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist_kuzushiji10 data: 
    `root/mnist_kuzushiji10/train/0/xx.png`
    `root/mnist_kuzushiji10/train/2/xx.png`
    `root/mnist_kuzushiji10/train/6/xx.png`
    `root/mnist_kuzushiji10/test/0/xx.png`
    `root/mnist_kuzushiji10/test/2/xx.png`
    `root/mnist_kuzushiji10/test/6/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist_kuzushiji10`,
              root should be `/user/.../mydata`.
        dataset: whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist_kuzushiji10`.
    """
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist_kuzushiji10')
    p.add(1)
    for url in Param.mnist_kuzushiji10:
        get_file(url, gfile.path_join(task_path, url.split('/')[-1]), verbose=0)
        p.add(1)
    train = np.load(gfile.path_join(task_path, 'kmnist-train-imgs.npz'))['arr_0']
    train_label = np.load(gfile.path_join(task_path, 'kmnist-train-labels.npz'))['arr_0']
    test = np.load(gfile.path_join(task_path, 'kmnist-test-imgs.npz'))['arr_0']
    test_label = np.load(gfile.path_join(task_path, 'kmnist-test-labels.npz'))['arr_0']
    p.add(1)
    for i in set(train_label):
        gfile.makedirs(gfile.path_join(task_path, 'train', str(i)))
    for i in set(test_label):
        gfile.makedirs(gfile.path_join(task_path, 'test', str(i)))
    p.add(1)
    for idx in range(train.shape[0]):
        save_image(gfile.path_join(task_path, 'train', str(train_label[idx]), str(idx)+'.png'), 
                   array_to_image(train[idx].reshape(28, 28, 1)))
    p.add(1)
    for idx in range(test.shape[0]):
        save_image(gfile.path_join(task_path, 'test', str(test_label[idx]), str(idx)+'.png'), 
                   array_to_image(test[idx].reshape(28, 28, 1)))
    p.add(1)
    for url in Param.mnist_kuzushiji10:
        gfile.remove(gfile.path_join(task_path, url.split('/')[-1]))
    p.add(1)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
                .join({'test':from_class_folder(gfile.path_join(task_path, 'test'), label_encoder=1)}))
    return task_path


def mnist_kuzushiji49(root=None, dataset=True, verbose=1):
    """Kuzushiji-49 from https://github.com/rois-codh/kmnist.
    
    Kuzushiji-49, as the name suggests, has 49 classes (28x28 grayscale, 270,912 images),
    is a much larger, but imbalanced dataset containing 48 Hiragana 
    characters and one Hiragana iteration mark.
    
    Each sample is an gray image (in 3D NDArray) with shape (28, 28, 1).
    
    Attention: if exist dirs `root/mnist_kuzushiji49`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist_kuzushiji49 data: 
    `root/mnist_kuzushiji49/train/0/xx.png`
    `root/mnist_kuzushiji49/train/2/xx.png`
    `root/mnist_kuzushiji49/train/6/xx.png`
    `root/mnist_kuzushiji49/test/0/xx.png`
    `root/mnist_kuzushiji49/test/2/xx.png`
    `root/mnist_kuzushiji49/test/6/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist_kuzushiji49`,
              root should be `/user/.../mydata`.
        dataset: whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist_kuzushiji49`.
    """
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist_kuzushiji49')
    p.add(1)
    for url in Param.mnist_kuzushiji49:
        get_file(url, gfile.path_join(task_path, url.split('/')[-1]), verbose=0)
        p.add(1)
    train = np.load(gfile.path_join(task_path, 'k49-train-imgs.npz'))['arr_0']
    train_label = np.load(gfile.path_join(task_path, 'k49-train-labels.npz'))['arr_0']
    test = np.load(gfile.path_join(task_path, 'k49-test-imgs.npz'))['arr_0']
    test_label = np.load(gfile.path_join(task_path, 'k49-test-labels.npz'))['arr_0']
    p.add(1)
    for i in set(train_label):
        gfile.makedirs(gfile.path_join(task_path, 'train', str(i)))
    for i in set(test_label):
        gfile.makedirs(gfile.path_join(task_path, 'test', str(i)))
    p.add(1)
    for idx in range(train.shape[0]):
        save_image(gfile.path_join(task_path, 'train', str(train_label[idx]), str(idx)+'.png'), 
                   array_to_image(train[idx].reshape(28, 28, 1)))
    p.add(1)
    for idx in range(test.shape[0]):
        save_image(gfile.path_join(task_path, 'test', str(test_label[idx]), str(idx)+'.png'), 
                   array_to_image(test[idx].reshape(28, 28, 1)))
    p.add(1)
    for url in Param.mnist_kuzushiji49:
        gfile.remove(gfile.path_join(task_path, url.split('/')[-1]))
    p.add(1)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
                .join({'test':from_class_folder(gfile.path_join(task_path, 'test'), label_encoder=1)}))
    return task_path


def mnist_kuzushiji_kanji(root=None, dataset=True, verbose=1):
    """Kuzushiji-Kanji dataset from https://github.com/rois-codh/kmnist.
    
    Kuzushiji-Kanji is a large and highly imbalanced 64x64 dataset 
    of 3832 Kanji characters, containing 140,426 images 
    of both common and rare characters.
    
    Attention: if exist dirs `root/mnist_kuzushiji_kanji`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist_kuzushiji_kanji data: 
    `root/mnist_kuzushiji_kanji/train/U+55C7/xx.png`
    `root/mnist_kuzushiji_kanji/train/U+7F8E/xx.png`
    `root/mnist_kuzushiji_kanji/train/U+9593/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist_kuzushiji_kanji`,
              root should be `/user/.../mydata`.
        dataset: whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist_kuzushiji_kanji.
    """
    if root is None:
        root = './'
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist_kuzushiji_kanji', make_root_dir=False)
    p.add(1)
    get_file(Param.mnist_kuzushiji_kanji[0], gfile.path_join(root, url.split('/')[-1]), verbose=0)
    p.add(7)
    decompress(gfile.path_join(root, url.split('/')[-1]), task_path)
    p.add(1)
    gfile.rename(gfile.path_join(task_path, 'kkanji2'), gfile.path_join(task_path, 'train'))
    gfile.remove(gfile.path_join(root, 'kkanji.tar'))
    p.add(1)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
               )
    return task_path


def mnist_tibetan(root=None, dataset=True, verbose=1):
    """Tibetan-MNIST from https://github.com/bat67/TibetanMNIST.
    
    Tibetan-MNIST is a drop-in replacement for the
    MNIST dataset (28x28 grayscale, 70,000 images), 
    provided in the original MNIST format as well as a NumPy format.
    Since MNIST restricts us to 10 classes, we chose one character to
    represent each of the 10 rows of Hiragana when creating Tibetan-MNIST.
    
    Each sample is an gray image (in 3D NDArray) with shape (28, 28, 1).
    
    Attention: if exist dirs `root/mnist_tibetan`, api will delete it and create it.
    Data storage directory:
    root = `/user/.../mydata`
    mnist_tibetan data: 
    `root/mnist_tibetan/train/0/xx.png`
    `root/mnist_tibetan/train/2/xx.png`
    `root/mnist_tibetan/train/6/xx.png`
    `root/mnist_tibetan/test/0/xx.png`
    `root/mnist_tibetan/test/2/xx.png`
    `root/mnist_tibetan/test/6/xx.png`
    Args:
        root: str, Store the absolute path of the data directory.
              example:if you want data path is `/user/.../mydata/mnist_tibetan`,
              root should be `/user/.../mydata`.
        dataset: whether to return a la.data.Dataset object.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
    Returns:
        Store the absolute path of the data directory, is `root/mnist_tibetan`.
    """
    p = Progbar(10, verbose=verbose)
    task_path = assert_dirs(root, 'mnist_tibetan')
    p.add(1)
    data = pd.DataFrame()
    for url in Param.mnist_tibetan:
        s = requests.get(url).content
        data = pd.concat([data, pd.read_csv(io.StringIO(s.decode('utf-8')), header=None, dtype='uint8')])
        p.add(3)
    train = data.loc[:, 1:].values.reshape(-1, 28, 28)
    train_label = data.loc[:, 0].values
    for i in set(train_label):
        gfile.makedirs(gfile.path_join(task_path, 'train', str(i)))
    p.add(1)
    for idx in range(train.shape[0]):
        save_image(gfile.path_join(task_path, 'train', str(train_label[idx]), str(idx)+'.png'),
                   array_to_image(train[idx].reshape(28, 28, 1)))
    p.add(2)
    if dataset:
        return (from_class_folder(gfile.path_join(task_path, 'train'), label_encoder=1).split({'train':1})
                .join({'test':from_class_folder(gfile.path_join(task_path, 'test'), label_encoder=1)}))
    return task_path


from linora.feature_column._numerical import *

class FeatureNormalize(object):
    def __init__(self):
        self.pipe = {}
        
    def normalize_max(self, variable, name=None, keep=True):
        """normalize feature with max method.

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'normalize_max', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_maxabs(self, variable, name=None, keep=True):
        """normalize feature with maxabs method.

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'normalize_maxabs', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_l1(self, variable, name=None, keep=True):
        """normalize feature with l1 method.

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'normalize_l1', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_l2(self, variable, name=None, keep=True):
        """normalize feature with l2 method.

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'normalize_l2', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_meanminmax(self, variable, name=None, keep=True):
        """normalize feature with meanminmax method.

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'normalize_meanminmax', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_minmax(self, variable, feature_range=(0, 1), name=None, keep=True):
        """normalize feature with minmax method.

        Args:
            variable: str, feature variable name.
            feature_range: list or tuple, range of values after feature transformation.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'feature_range':feature_range, 'name':variable if name is None else name},
                  'type':'normalize_minmax', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_norm(self, variable, name=None, keep=True):
        """normalize feature with norm method.

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'normalize_norm', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def normalize_robust(self, variable, feature_scale=(0.5, 0.5), name=None, keep=True):
        """normalize feature with robust method.

        Args:
            variable: str, feature variable name.
            feature_scale: list or tuple, each element is in the [0,1] interval.
                (feature_scale[0], feature.quantile(0.5+feature_scale[1]/2)-feature.quantile(0.5-feature_scale[1]/2)).
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'feature_scale':feature_scale, 'name':variable if name is None else name},
                  'type':'normalize_robust', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    
    
    
        


from linora.image._image_util import list_images

__all__ = ['list_audios']


def list_audios(directory, file_format=('mp3', 'flac', 'wav', 'ogg', 'ape')):
    """Lists all audios in a directory, including all subdirectories.
    
    Args:
        directory: string, absolute path to the directory.
        file_format: tuple of strings or single string, extensions of the audios.
    Returns:
        a list of audio paths.
    """
    return list_images(directory, file_format=file_format)


import numpy as np

class Vlines():
    def add_vlines(self, name, xdata, ymin, ymax, **kwargs):
        """Plot vertical lines at each *x* from *ymin* to *ymax*.
        
        Args:
            name: data name.
            xdata: x-axis data, float or array-like.
            ymin, ymax : float or array-like
                Respective beginning and end of each line. If scalars are
                provided, all lines will have same length.
            colors : list of colors, default: :rc:`lines.color`
            linestyles : {'solid', 'dashed', 'dashdot', 'dotted'}, optional
            label : str, default: ''
        """
        if isinstance(xdata, (int, float)):
            xdata = [xdata]
        if 'color' not in kwargs and 'colors' not in kwargs and 'linecolor' not in kwargs:
            kwargs['colors'] = [self._params.color.pop(0)[1]]*len(xdata)
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['ymin'] = ymin
        self._params.ydata[name]['ymax'] = ymax
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['plotmode'] = 'vlines'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_vlines
        return self
    
    def _execute_plot_vlines(self, fig, ax, i, j):
        ax_plot = ax.vlines(j['xdata'], j['ymin'], j['ymax'], **j['kwargs'])


import random
import functools
import itertools
from collections import defaultdict

import numpy as np
import pandas as pd

from linora import gfile
from linora.utils._config import Config

__all__ = ['from_tensor']


class DataSet():
    def __init__(self):
        self._params_init()
    
    def _params_init(self):
        self._params = Config()
        self._params.batch = 0
        self._params.batch_size = 1
        self._params.skip_size = None
        self._params.take_size = -1
        self._params.shuffle_size = 1
        self._params.prefetch_size = 1
        self._params.sample = 0
        self._params.step = 1
        self._params.tensor_mode = 'numpy'
        self._params.options = defaultdict(dict)
        
    def batch(self, batch_size, drop_remainder=False):
        """Combines consecutive elements of this dataset into batches.
        
        Args:
            batch_size: representing the number of consecutive elements of this dataset to combine in a single batch.
            drop_remainder: representing whether the last batch should be dropped in the case it has fewer than batch_size elements; 
                            the default behavior is not to drop the smaller batch.
        """
        assert 'batch' not in self._params.options, '`batch` already exists.'
        assert isinstance(batch_size, int) and batch_size>0, '`batch_size` type should be int and greater than 0.'
        self._params.batch_size = batch_size
        self._params.drop_remainder = drop_remainder
        self._params.options['batch'].update({self._params.step: {'batch_size':batch_size, 'drop_remainder':drop_remainder}})
        self._params.step += 1
        return self
    
    def concatenate(self, dataset):
        """Creates a Dataset by concatenating the given dataset with this dataset.
        
        Args:
            dataset: la.data.TextLineDataset object, Dataset to be concatenated.
        """
        self._params.data.append(dataset._params.data)
        self._params.options['concatenate'].update({self._params.step: None})
        self._params.step += 1
        return self
        
    def enumerate(self, start=0):
        """Enumerates the elements of this dataset.
        
        Args:
            start: int, representing the start value for enumeration.
        """
        assert 'enumerate' not in self._params.options, '`enumerate` already exists.'
        self._params.enumerate = start
        self._params.options['enumerate'].update({self._params.step: {'start':start}})
        self._params.step += 1
        return self
    
#     def filter(self, filter_func):
#         """A transformation that filter dataset based on a filter_func.
        
#         Args:
#             filter_func: A function that return True or False
#         """
#         if self.params.data_mode=='list':
#             filter_list = [i for i in range(len(self.params.data[0])) if filter_func([j[i] for j in self.params.data])]
#         else:
#             filter_list = [r for r, i in enumerate(self.params.data) if filter_func(i)]
#         if filter_list:
#             self.params.data_index = [i for i in self.params.data_index if i not in filter_list]
#         self.params.options['filter'].append((self.params.step, filter_func))
#         self.params.step += 1
#         return self
    
    def map(self, map_func):
        """Maps map_func across the elements of this dataset.
        
        Args:
            map_func: A function mapping a dataset element to another dataset element.
            map_size: representing the number elements to process asynchronously in parallel. 
        """
        assert 'map' not in self._params.options, '`map` already exists.'
        self._params.map_func = map_func
        self._params.options['map'].update({self._params.step: {'map_func':map_func}})
        self._params.step += 1
        return self
    
    def options(self):
        """Returns the options for this dataset and its inputs."""
        return self._params.options
    
    def prefetch(self, prefetch_size):
        """Creates a Dataset that prefetches elements from this dataset.
        
        Args:
            prefetch_size: representing the maximum number of elements that will be buffered when prefetching.
        """
        assert 'prefetch' not in self._params.options, '`prefetch` already exists.'
        assert 'take_while' not in self._params.options, '`prefetch` must be placed in `take_while` front.'
        assert isinstance(prefetch_size, int) and prefetch_size>0, '`prefetch_size` type should be int and greater than 0.'
        self._params.prefetch_size = prefetch_size
        self._params.options['prefetch'].update({self._params.step: {'prefetch_size':prefetch_size}})
        self._params.step += 1
        return self
        
#     def reduce(self, reduce_func):
#         """Reduces the input dataset to a single element.
        
#         Args:
#             reduce_func: A function that maps to new_state. It must take two arguments and return a new element
#         """
#         if self.params.data_mode=='list':
#             return [functools.reduce(reduce_func, i[self.params.data_index]) for i in self.params.data]
#         return functools.reduce(reduce_func, self.params.data[self.params.data_index])
    
    def repeat(self, repeat_size):
        """Repeats this dataset so each original value is seen count times.
        
        Args:
            repeat_size: representing the number of times the dataset should be repeated.
        """
        assert 'take_while' not in self._params.options, '`repeat` must be placed in `take_while` front.'
        assert isinstance(repeat_size, int) and repeat_size>0, '`repeat_size` type should be int and greater than 0.'
        self._params.data = self._params.data*(repeat_size+1)
        self._params.options['repeat'].update({self._params.step: {'repeat_size':repeat_size}})
        self._params.step += 1
        return self
        
#     def shard(self, shard_size, shard_index):
#         """Creates a Dataset that includes only 1/num_shards of this dataset.
        
#         Args:
#             shard_size: representing the number of shards operating in parallel.
#             shard_index: representing the worker index.
#         """
#         assert 'take_while' not in self.params.options, '`shard` must be placed in `take_while` front.'
#         assert isinstance(shard_size, int) and shard_size>0, '`shard_size` type should be int and greater than 0.'
#         assert isinstance(shard_index, int) and shard_index>=0, '`shard_index` type should be int and greater than or equal to 0.'
#         self.params.data_index = [self.params.data_index[i] for i in range(shard_index, len(self.params.data_index), shard_size)]
#         self.params.options['shard'].append((self.params.step, shard_size, shard_index))
#         self.params.step += 1
#         return self
    
    def shuffle(self, shuffle_size, seed=None):
        """Randomly shuffles the elements of this dataset.
        
        Args:
            shuffle_size: representing the number of elements from this dataset from which the new dataset will sample.
            seed: representing the random seed that will be used to create the distribution.
        """
        assert 'shuffle' not in self._params.options, '`shuffle` already exists.'
        assert 'take_while' not in self._params.options, '`shuffle` must be placed in `take_while` front.'
        assert isinstance(shuffle_size, int) and shuffle_size>-2 and shuffle_size!=0, '`shuffle_size` type should be int and greater than 0 or equal to -1.'
        self._params.shuffle_size = shuffle_size
        self._params.options['shuffle'].update({self._params.step: {'shuffle_size':shuffle_size, 'seed':seed}})
        self._params.step += 1
        return self
    
    def skip(self, skip_size):
        """Creates a Dataset that skips count elements from this dataset.
        
        Skip all data for the first file at most.
        
        Args:
            skip_size: representing the number of elements of this dataset that should be skipped to form the new dataset. 
                       If count is greater than the size of this dataset, the new dataset will contain no elements.
        """
        assert 'skip' not in self._params.options, '`skip` already exists.'
        assert 'take_while' not in self._params.options, '`skip` must be placed in `take_while` front.'
        assert isinstance(skip_size, int) and skip_size>0, '`skip_size` type should be int and greater than 0.'
        self._params.skip_size = skip_size
        self._params.options['skip'].update({self._params.step: {'skip_size':skip_size}})
        self._params.step += 1
        return self
        
    def take(self, take_size):
        """Creates a Dataset with at most count elements from this dataset.
        
        Args:
            take_size: representing the number of elements of this dataset that should be taken to form the new dataset. 
                       If count is -1, or if count is greater than the size of this dataset, 
                       the new dataset will contain all elements of this dataset.
        """
        assert 'take' not in self._params.options, '`take` already exists.'
        assert 'take_while' not in self._params.options, '`take` must be placed in `take_while` front.'
        assert isinstance(take_size, int) and take_size>-2 and take_size!=0, '`take_size` type should be int and greater than 0 or equal to -1.'
        self._params.take_size = take_size
        self._params.options['take'].update({self._params.step: {'take_size':take_size}})
        self._params.step += 1
        return self
    
#     def take_while(self, take_func):
#         """A transformation that stops dataset iteration based on a take_func.
        
#         Args:
#             take_func: A function that return True or False
#         """
#         temp = set()
#         index = self.params.data_index[:max([self.params.data_index.index(i) for i in range(len(self.params.data))])+1]
#         for r, i in enumerate(index):
#             if i in temp:
#                 continue
#             temp.add(i)
#             if self.params.data_mode=='list':
#                 if take_func([j[i] for j in self.params.data]):
#                     self.params.data_index = self.params.data_index[:r]
#                     break
#             else:
#                 if take_func(self.params.data[i]):
#                     self.params.data_index = self.params.data_index[:r]
#                     break
#         self.params.options['take_while'].append((self.params.step, take_func))
#         self.params.step += 1
#         return self

    def to_tensor(self, mode='tf'):
        """Transform data from numpy array to tensor.
        
        Args:
            mode: Deep learning framework name, one of ['tf', 'pytorch', 'paddle', 'mxnet', 'mindspore'].
        """
        assert 'to_tensor' not in self._params.options, '`to_tensor` already exists.'
        assert 'take_while' not in self._params.options, '`take` must be placed in `take_while` front.'
        if mode in ['tf', 'tensorflow']:
            from tensorflow import convert_to_tensor
            self._params.framework = convert_to_tensor
        elif mode in ['pytorch', 'torch']:
            from torch import as_tensor
            self._params.framework = as_tensor
        elif mode in ['paddle', 'paddlepaddle']:
            from paddle import to_tensor
            self._params.framework = to_tensor
        elif mode in ['mx', 'mxnet']:
            from mxnet.ndarray import array
            self._params.framework = array
        elif mode in ['mindspore']:
            from mindspore.numpy import array
            self._params.framework = array
        else:
            raise ValueError('`mode` value error.')
        self._params.tensor_mode = mode
        self._params.options['to_tensor'].update({self._params.step: {'mode':mode}})
        self._params.step += 1
        return self
    
    def _to_tensor(self, data):
        if self._params.tensor_mode=='numpy':
            return data
        return self._params.framework(data)
    
    def __iter__(self):
        self._params.shuffle_size = np.ceil(max(self._params.shuffle_size, self._params.prefetch_size, 1)/self._params.batch_size)*self._params.batch_size
        self._params.df = pd.read_csv(self._params.data[self._params.batch_file], sep=self._params.sep, 
                                     iterator=True, header=self._params.header, skiprows=self._params.skip_size)
        self._params.values = self._params.df.get_chunk(self._params.shuffle_size)
        if 'shuffle' in self._params.options:
            self._params.values = self._params.values.sample(frac=1, random_state=self._params.shuffle_seed).reset_index(drop=True)
        self._params.batch_file += 1
        self._params.batch_index = 0
        return self
        
    def __next__(self):
        values = self._params.values.loc[self._params.batch_size*self._params.batch_index:self._params.batch_size*(self._params.batch_index+1)]
        if len(values)<self._params.batch_size:
            self._params.batch_index = 0
            try:
                self._params.values = self._params.df.get_chunk(self._params.shuffle_size).reset_index(drop=True)
            except StopIteration:
                if self._params.batch_file==len(self._params.data):
                    raise StopIteration
                self._params.df = pd.read_csv(self._params.data[self._params.batch_file], sep=self._params.sep, header=self._params.header, iterator=True)
                self._params.values = self._params.df.get_chunk(self._params.shuffle_size)
                self._params.batch_file += 1
            if 'shuffle' in self._params.options:
                self._params.values = self._params.values.sample(frac=1, random_state=self._params.shuffle_seed).reset_index(drop=True)
            values = self._params.values.loc[0:self._params.batch_size]
            
        self._params.batch += 1
        self._params.batch_index += 1
        if self._params.take_size>0:
            if self._params.sample>=self._params.take_size:
                raise StopIteration
            self._params.sample += len(values)
        if 'map' in self._params.options:
            return self._to_tensor(values.apply(self._params.map_func, axis=1).values)
        return self._to_tensor(values.values)


class from_tensor(DataSet):
    """Creates a Dataset comprising lines from one or more text files.

    Args:
        data: file path string or list of file path.
        sep: str, default ',', Delimiter to use.
        header: int, list of int, Row number(s) to use as the column names, and the start of the data.
    """
    def __init__(self, data, sep=',', header=None):
        super(from_tensor, self).__init__()
        self._params.rank_list = None
        self._params.batch_index = 0
        self._params.batch_file = 0
        self._params.sample = 0
        self._params.sep = sep
        self._params.header = header
        data = list(data) if isinstance(data, (tuple, list)) else [data]
        for i in data:
            if not gfile.isfile(i):
                raise ValueError(f'`{i}` not a file path.')
        self._params.data = data
        return self


import numpy as np
from PIL import Image

__all__ = ['crop', 'crop_central', 'crop_point', 'concat', 'split', 'paste']


def crop(image, box, p=1):
    """Returns a rectangular region from this image. 
    
    The box is a 4-tuple defining the left, upper, right, and lower pixel coordinate.
    Args:
        image: a PIL instance.
        box: The crop rectangle, as a (left, upper, right, lower)-tuple.
             if [x,y], transform [x,x,y,y] 4-tuple.
        p: probability that the image does this. Default value is 1.
    returns: 
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if len(box)==2:
        if isinstance(box[0], (list, tuple)):
            box = [box[0][0], box[0][1], box[1][0], box[1][1]]
        else:
            box = [box[0], box[0], box[1], box[1]]
    return image.crop(box)


def crop_central(image, central_rate, p=1):
    """Crop the central region of the image.
    
    Remove the outer parts of an image but retain the central region of the image
    along each dimension. If we specify central_fraction = 0.5, this function
    returns the region marked with "X" in the below diagram.
       --------
      |        |
      |  XXXX  |
      |  XXXX  |
      |        |   where "X" is the central 50% of the image.
       --------
    
    Args:
        image: a PIL instance.
        central_rate: if int float, should be in the interval (0, 1], fraction of size to crop.
                      if tuple list, randomly picked in the interval
                      `[central_rate[0], central_rate[1])`, value is fraction of size to crop.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    Raises:
        ValueError: if central_crop_fraction is not within (0, 1].
    """
    if np.random.uniform()>p:
        return image
    if isinstance(central_rate, (int, float)):
        assert 0<central_rate<=1, 'if central_rate type one of int or float, must be in the interval (0, 1].'
    elif isinstance(central_rate, (tuple, list)):
        assert 0<central_rate[0]<central_rate[1]<=1, 'central_rate should be 1 >= central_rate[1] > central_rate[0] > 0.'
        central_rate = np.random.uniform(central_rate[0], central_rate[1])
    else:
        raise ValueError('central_rate should be one of int, float, tuple, list.')
    left = int(image.size[0]*(0.5-central_rate/2))
    upper = int(image.size[1]*(0.5-central_rate/2))
    right = int(image.size[0]*(0.5+central_rate/2))
    lower = int(image.size[1]*(0.5+central_rate/2))
    return image.crop((left, upper, right, lower))


def crop_point(image, height_rate, width_rate, p=1):
    """Crop the any region of the image.
    
    Crop region area = height_rate * width_rate *image_height * image_width
    
    Args:
        image: a PIL instance.
        height_rate: float, in the interval (0, 1].
        width_rate: float, in the interval (0, 1].
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    Raises:
        ValueError: if central_crop_fraction is not within (0, 1].
    """
    if np.random.uniform()>p:
        return image
    assert 0<height_rate<=1 and 0<width_rate<=1, 'height_rate and width_rate should be in the interval (0, 1].'
    left = image.size[0]*np.random.uniform(0, 1-width_rate)
    upper = image.size[1]*np.random.uniform(0, 1-height_rate)
    right = left+image.size[0]*width_rate
    lower = upper+image.size[1]*height_rate
    return image.crop((left, upper, right, lower))


def paste(image1, image2, box=None):
    """Pastes another image into this image. 
    
    Args:
        image1: a PIL instance. The first image.
        image2: a PIL instance. The second image.
        box: An optional 4-tuple giving the region to paste into. 
             If a 2-tuple is used instead, it’s treated as the upper left corner. 
             If omitted or None, the source is pasted into the upper left corner.
    Return:
        a PIL instance.
    """
    image = image1.copy()
    if box is None:
        box = [0, 0, min(image2.size[0], image.size[0]), min(image2.size[1], image.size[1])]
    elif len(box)==4:
        box = [box[0], box[1], min(box[2], image.size[0]), min(box[3], image.size[1])]
    elif len(box)==2:
        box = [box[0], box[1], min(box[0]+image2.size[0], image.size[0]), min(box[1]+image2.size[1], image.size[1])]
    else:
        raise ValueError('`box` value error.')
    if (box[2]-box[0])!=image2.size[0] or (box[3]-box[1])!=image2.size[1]:
        image.paste(image2.resize([box[2]-box[0], box[3]-box[1]], resample=Image.Resampling.NEAREST), box)
    else:
        image.paste(image2, box)
    return image


def concat(image, col_nums=None):
    """Merge multiple pictures into one picture.
    
    Splice multiple pictures from left to right and from top to bottom to one picture.
    
    Args:
        image: list or tuple or dict, a list or tuple of PIL instance.
               if dict, should be la.image.split function output.
        col_nums: int, maximum number of columns spliced.
    returns:
        a PIL instance.
    """
    if isinstance(image, dict):
        if image['image_strides'][0]>image['image_size'][0] or image['image_strides'][1]>image['image_size'][1]:
            image = image['image_list']
        else:
            img = Image.new(image['image_list'][0].mode, image['image_shape'], (0,0,0))
            for r, i in enumerate(image['image_true']):
                img.paste(image['image_list'][r].crop(i[1]), i[0])
            return img
    if isinstance(image, (list, tuple)):
        image = list(image)
        if len(image)==1:
            return image[0]
        size = image[0].size
        for r, i in enumerate(image):
            if size!=i.size:
                image[r] = i.resize(size)
        if col_nums is None or col_nums<1:
            col_nums = np.ceil(np.sqrt(len(image)))
        shape = [int(np.ceil(len(image)/col_nums)), int(col_nums)]
        img = Image.new(image[0].mode, [size[0]*shape[1], size[1]*shape[0]], (0,0,0))
        for r, i in enumerate(image):
            img.paste(i, (r%shape[1]*size[0], r//shape[1]*size[1]))
        return img
    raise ValueError('`image` should be list or tuple or dict.')
    

def split(image, size, strides=None, keep_last=True):
    """Split one picture into multiple pictures.
    
    Args:
        image: a PIL instance.
        size: list or tuple, split picture size, [width, height].
        strides: int or tuple or list, eg. [width, height], specifying the crop strides along the height and width. 
        keep_last: bool, whether to preserve the clipped image when the right and bottom edges are not long enough.
    returns:
        a dict.
    """
    if strides is None:
        strides = size
    elif isinstance(strides, int):
        strides = [strides, strides]
    shape = image.size
    assert shape[0]>size[0] and shape[1]>size[1] and shape[0]>strides[0] and shape[1]>strides[1], '`size` or `strides` value error.'
    xaxis = []
    yaxis = []
    xtrue = []
    ytrue = []
    i = 0
    while 1:
        if i+size[0]==shape[0]:
            xaxis.append(i)
            xtrue.append([[i, i+size[0]], [0, size[0]]])
            break
        elif i+size[0]<shape[0]:
            xaxis.append(i)
            xtrue.append([[i, i+strides[0]], [0, strides[0]]])
            if i+size[0]+strides[0]>shape[0]:
                if keep_last:
                    xaxis.append(shape[0]-size[0])
                    xtrue.append([[i+strides[0], shape[0]], [size[0]-(shape[0]-i-strides[0]), size[0]]])
                break
        i += strides[0]
    i = 0
    while 1:
        if i+size[1]==shape[1]:
            yaxis.append(i)
            ytrue.append([[i, i+size[1]], [0, size[1]]])
            break
        elif i+size[1]<shape[1]:
            yaxis.append(i)
            ytrue.append([[i, i+strides[1]], [0, strides[1]]])
            if i+size[1]+strides[1]>shape[1]:
                if keep_last:
                    yaxis.append(shape[1]-size[1])
                    ytrue.append([[i+strides[1], shape[1]], [size[1]-(shape[1]-i-strides[1]), size[1]]])
                break
        i += strides[1]
    if not keep_last:
        shape = [xaxis[-1]+strides[0], yaxis[-1]+strides[1]]
    return {'image_list': [image.crop([j, i, j+size[0], i+size[1]]) for i in yaxis for j in xaxis],
             'image_size': size, 'image_strides': strides, 'image_block':[len(xaxis), len(yaxis)], 'image_shape':shape,
             'image_true': [[[j[0][0], i[0][0], j[0][1], i[0][1]], [j[1][0], i[1][0], j[1][1], i[1][1]]] for i in ytrue for j in xtrue]}


import json
import matplotlib.pyplot as plt

from linora.utils._config import Config

__all__ = ['Grid']


class Grid():
    def __init__(self, nrows, ncols, left=None, bottom=None, right=None, top=None,
                 wspace=None, hspace=None, width_ratios=None, height_ratios=None):
        """A grid layout to place subplots within a figure.
        
        Args:
            nrows, ncols : int
                The number of rows and columns of the grid.
                
            left, right, top, bottom : float, optional
                Extent of the subplots as a fraction of figure width or height.
                Left cannot be larger than right, and bottom cannot be larger than
                top. If not given, the values will be inferred from a figure or
                rcParams at draw time.

            wspace : float, optional
                The amount of width reserved for space between subplots,
                expressed as a fraction of the average axis width.
                If not given, the values will be inferred from a figure or
                rcParams when necessary.

            hspace : float, optional
                The amount of height reserved for space between subplots,
                expressed as a fraction of the average axis height.
                If not given, the values will be inferred from a figure or
                rcParams when necessary.

            width_ratios : array-like of length *ncols*, optional
                Defines the relative widths of the columns. Each column gets a
                relative width of ``width_ratios[i] / sum(width_ratios)``.
                If not given, all columns will have the same width.

            height_ratios : array-like of length *nrows*, optional
                Defines the relative heights of the rows. Each column gets a
                relative height of ``height_ratios[i] / sum(height_ratios)``.
                If not given, all rows will have the same height.
        """
        self._grid = Config()
        self._grid.figure = {'figsize':(12.8, 7.2)}
        self._grid.grid = {'nrows':nrows, 'ncols':ncols, 'left':left, 'bottom':bottom, 
                           'right':right, 'top':top, 'wspace':wspace, 'hspace':hspace, 
                           'width_ratios':width_ratios, 'height_ratios':height_ratios}
        self._grid.grid_id = dict()
        self._grid.title = {'title':None}
        
    def add_plot(self, grid_id, plot):
        """Add la.chart.Plot object in designated area.
        
        Args:
            grid_id: str, grid area.
            plot: a la.chart.Plot object.
        """
        grid_id = [i.strip() for i in grid_id.split(',')]
        assert len(grid_id)==2, '`grid_id` value error.'
        for i in range(2):
            if ':' in grid_id[i]:
                t = grid_id[i].split(':')
                if t[0]=='':
                    t[0] = '0'
                if t[1]=='':
                    t[1] = self._grid.grid['nrows'] if i==0 else self._grid.grid['ncols']
                grid_id[i] = [int(t[0]), int(t[1])]
            else:
                if '-' in grid_id[i]:
                    if i==0:
                        grid_id[i] = [self._grid.grid['nrows']+int(grid_id[i]), self._grid.grid['nrows']+int(grid_id[i])+1]
                    else:
                        grid_id[i] = [self._grid.grid['ncols']+int(grid_id[i]), self._grid.grid['ncols']+int(grid_id[i])+1]
                else:
                    grid_id[i] = [int(grid_id[i]), int(grid_id[i])+1]
        
        self._grid.grid_id[len(self._grid.grid_id)] = {'grid_id':grid_id, 'plot':plot}        
        return self
    
    def get_config(self, json_path=None):
        config = {
            'mode': 'grid',
            'grid': self._grid.grid,
            'grid_id': self._grid.grid_id,
            'figure': self._grid.figure,
            'title': self._grid.title
        }
        if json_path is not None:
            with open(json_path, 'w') as f:
                json.dump(config, f)
        return config
    
    def set_config(self, config):
        if isinstance(config, str):
            assert config.endswith('.json'), f'{config} not a json file.'
            with open(config) as f:
                config = json.load(f)
        assert isinstance(config, dict), f'{config} not a dict.'
        assert config['mode']=='grid', 'config info not match.'
        self._grid.figure = config['figure']
        self._grid.grid = config['grid']
        self._grid.grid_id = config['grid_id']
        self._grid.title = config['title']
        return self
        
    def render(self, image_path=None, if_show=True, **kwargs):
        """show and save plot."""
        fig = self._execute()
        if image_path is not None:
            fig.savefig(image_path, **kwargs)
        if if_show:
            return fig.show()
    
    def set_figure(self, width=12.8, height=7.2, dpi=None, facecolor=None, edgecolor=None, frameon=True, clear=False):
        """Add figure config.
        
        Args:
            width: float, figure size width in inches.
            height: float, figure size height in inches.
            dpi: float, The resolution of the figure in dots-per-inch.
            facecolor: color, The background color.
            edgecolor: color, The border color.
            frameon: bool, default: True, If False, suppress drawing the figure frame.
            clear: bool, default: False, If True and the figure already exists, then it is cleared.
        """
        kwargs = {'figsize':(width, height), 'dpi':dpi, 'facecolor':facecolor, 
                  'edgecolor':edgecolor, 'frameon':frameon, 'clear':clear}
        self._grid.figure.update(kwargs)
        return self
    
    def set_title(self, title, x=0.5, y=0.98, horizontalalignment='center', 
                  verticalalignment='top', fontsize='large', fontweight='normal', **kwargs):
        """Add a centered suptitle to the figure.
        
        Args:
            title : str
                The suptitle text.
            x : float, default: 0.5
                The x location of the text in figure coordinates.
            y : float, default: 0.98
                The y location of the text in figure coordinates.
            horizontalalignment, ha : {'center', 'left', 'right'}, default: center
                The horizontal alignment of the text relative to (*x*, *y*).
            verticalalignment, va : {'top', 'center', 'bottom', 'baseline'}, default: top
                The vertical alignment of the text relative to (*x*, *y*).
            fontsize, size : default: 'large'
                The font size of the text. 
            fontweight, weight : default: 'normal'
                The font weight of the text. 
        """
        kwargs['t'] = title
        kwargs['x'] = x
        kwargs['y'] = y
        kwargs['horizontalalignment'] = horizontalalignment
        kwargs['verticalalignment'] = verticalalignment
        kwargs['fontsize'] = fontsize
        kwargs['fontweight'] = fontweight
        self._grid.title.update(kwargs)
        return self
    
    def _execute(self):
        fig = plt.figure(**self._grid.figure)
        grid = plt.GridSpec(**self._grid.grid)
        for i, grid_id in self._grid.grid_id.items():
            with plt.style.context(grid_id['plot']._params.theme):
                ax = fig.add_subplot(grid[grid_id['grid_id'][0][0]:grid_id['grid_id'][0][1], 
                                          grid_id['grid_id'][1][0]:grid_id['grid_id'][1][1]])
                ax = grid_id['plot']._execute_ax(fig, ax)
        if self._grid.title['title'] is not None:
            fig.suptitle(**self._grid.title)
        return fig


import numpy as np
from linora.utils._config import Config

__all__ = ['Chain', 'Sequential']


class Chain():
    """Chains list of learning rate schedulers. 

    Args:
        schedulers: list, List of chained schedulers.
    """
    def __init__(self, schedulers):
        self._params = Config()
        self._params.lr = schedulers[0]._params.lr
        self._params.schedulers = schedulers
        self._params.name = 'LRChain'
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        for schedulers in self._params.schedulers:
            schedulers._update(batch, log)
        self._params.lr = np.cumprod([i._params.lr for i in self._params.schedulers])[-1]
        

class Sequential():
    """Receives the list of schedulers that is expected to be called sequentially 
    during optimization process and milestone points that provides exact intervals 
    to reflect which scheduler is supposed to be called at a given batch.

    Args:
        schedulers: list, List of chained schedulers.
        batch_list: list, List of batch indices.
    """
    def __init__(self, schedulers, batch_list):
        self._params = Config()
        self._params.lr = schedulers[0]._params.lr
        self._params.schedulers = schedulers
        self._params.batch_list = sorted(batch_list)
        self._params.name = 'LRSequential'
    
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.batch_list:
            if batch>=self._params.batch_list[0]:
                self._params.batch_list.pop(0)
                self._params.schedulers.pop(0)
        self._params.schedulers[0]._update(batch, log)
        self._params.lr = self._params.schedulers[0]._params.lr
        


import itertools

import numpy as np
import pandas as pd

__all__ = ['sampling_random', 'sampling_systematic', 'sampling_stratify']


def sampling_random(feature, n=None, frac=None, replace=False, weights=None, seed=None):
    """simple random sampling.
    
    Args:
        feature: pd.DataFrame or pd.Series or np.array or list.
        n: int, optional, Number of items from axis to return. 
           Cannot be used with `frac`. Default = 1 if `frac` = None.
        frac: float, optional, Fraction of axis items to return. Cannot be used with `n`.
        replace: bool, default False, Allow or disallow sampling of the same row more than once.
        weights: str or pd.Series or np.array or list, optional, 
                 If passed a str, using a `feature` column as weights.
                 If passed a Series, will align with target object on index. 
                 Index values in weights not found in sampled object will be ignored and 
                 index values in sampled object not in weights will be assigned weights of zero.
                 If weights do not sum to 1, they will be normalized to sum to 1.
                 Missing values in the weights column will be treated as zero.
        seed: int, random seed.
    Returns:
        a feature index list of simple random sampling.
    """
    if not isinstance(feature, (pd.DataFrame, pd.Series)):
        feature = pd.Series(range(len(feature)))
    return feature.sample(n=n, frac=frac, replace=replace, weights=weights, random_state=seed).index.to_list()


def sampling_systematic(feature, n=None, frac=None, seed=None):
    """systematic sampling.
    
    Args:
        feature: pd.DataFrame or pd.Series or np.array or list.
        n: int, optional, Number of items from axis to return. 
           Cannot be used with `frac`. Default = 1 if `frac` = None.
        frac: float, optional, Fraction of axis items to return. Cannot be used with `n`.
        seed: int, random seed.
    Returns:
        a feature index list of simple systematic sampling.
    """
    if not isinstance(feature, (pd.DataFrame, pd.Series)):
        t = np.array(range(len(feature)))
    else:
        t = feature.index.to_numpy()
    if n is None and frac is None:
        raise ValueError('Only one of `n` and `frac` can be None.')
    elif n is not None and frac is not None:
        raise ValueError('Only one of `n` and `frac` can be None.')
    elif n is not None:
        frac = min(1, n/len(feature))
    else:
        frac = max(0, min(1, frac))
    n = int(frac*len(feature))
    interval = max(2, int(np.floor(1/frac)))
    seed = np.random.choice(range(interval), 1)[0] if seed is None else seed%interval
    index = list(range(seed, len(feature), interval))
    while len(index)>n:
        if (len(index)-n)%2:
            index.pop(int(len(index)*(seed%100/100)))
        else:
            index.pop(int(len(index)*(seed%10/10)))
    while len(index)<n:
        s = list(set(range(0, len(feature))).difference(set(index)))
        if (n-len(index))%2:
            index.append(s.pop(int(len(s)*(seed%100/100))))
        else:
            index.append(s.pop(int(len(s)*(seed%10/10))))
    return t[sorted(index)].tolist()


def _sampling_stratify(feature, n=None, frac=None, seed=None):
    var_numerical = [i for i in feature.columns if feature[i].nunique()>len(feature)*0.6 and feature[i].dtype.name[:3] in ['int', 'flo']]
    var_categorical = [i for i in feature.columns if i not in var_numerical]
    if len(var_categorical)>0:
        var_categorical = [j[0] for j in sorted([(i, feature[i].nunique()) for i in var_categorical], key=lambda x:x[1])]
    
    for r, i in enumerate(var_numerical):
        feature.loc[feature[i]<feature[i].median(), f'la_fea_{r}'] = 0
        feature.loc[feature[i]>=feature[i].median(), f'la_fea_{r}'] = 1
    var_categorical += [i for i in feature.columns if 'la_fea_' in i]
    
    sample = []
    group = feature[var_categorical].value_counts()
    if group[group>1].count()>0:
        for i in group[group>1].index:
            sample.append(feature.query('&'.join([str(m)+'=='+str(n) for m,n in zip(var_categorical, i)])).index.tolist())
    
    if group[group==1].count()>0:
        t = feature.query('('+')|('.join(['&'.join([str(m)+'=='+str(n) for m,n in zip(var_categorical, i)]) for i in group[group==1].index])+')')
        while len(t)>0:
            var_categorical = var_categorical[:-1]
            if len(var_categorical)==0:
                sample.append(t.index.tolist())
                break
            group = t[var_categorical].value_counts()
            if group[group>1].count()>0:
                for i in group[group>1].index:
                    sample.append(t.query('&'.join([str(m)+'=='+str(n) for m,n in zip(var_categorical, i)])).index.tolist())
            if group[group==1].count()==0:
                break
            t = t.query('('+')|('.join(['&'.join([str(m)+'=='+str(n) for m,n in zip(var_categorical, i)]) for i in group[group==1].index])+')')
    if n is not None:
        frac = n/len(feature)
    sample = [pd.Series(i).sample(n=max(1, int(np.ceil(len(i)*frac))), random_state=seed).to_list() for i in sample]
    sample = list(itertools.chain.from_iterable(sample))
    return sample[:n] if n is not None else sample[:int(np.ceil(len(feature)*frac))]


def sampling_stratify(feature, stratify=None, n=None, frac=None, seed=None):
    """stratify sampling.
    
    Args:
        feature: pd.DataFrame.
        stratify: pd.Series, shape (n_samples,), The target variable for supervised learning problems.
        n: int, optional, Number of items from axis to return. 
           Cannot be used with `frac`. Default = 1 if `frac` = None.
        frac: float, optional, Fraction of axis items to return. Cannot be used with `n`.
        seed: int, random seed.
    Returns:
        a feature index list of stratified stratify sampling.
    """
    if n is None and frac is None:
        raise ValueError('Only one of `n` and `frac` can be None.')
    elif n is not None and frac is not None:
        raise ValueError('Only one of `n` and `frac` can be None.')
    elif n is not None:
        frac = min(1, n/len(feature))
    else:
        frac = max(0, min(1, frac))
    
    if not isinstance(feature, pd.DataFrame):
        raise ValueError('feature type must be pd.DataFrame.')
    if stratify is not None:
        if not isinstance(stratify, pd.Series):
            raise ValueError('stratify type must be pd.Series.')
        if stratify.nunique()>len(stratify)*0.6 and stratify.dtype.name[:3] in ['int', 'flo']:
            return _sampling_stratify(feature, n=n, seed=seed)
        else:
            index = [_sampling_stratify(feature.loc[stratify==i].copy(), frac=frac, seed=seed) for i in stratify.unique()]
            index = list(itertools.chain.from_iterable(index))
            return index[:n] if n is not None else index[:int(np.ceil(len(feature)*frac))]
    else:
        index = _sampling_stratify(feature, frac=frac, seed=seed)
        return index[:n] if n is not None else index


import os

import numpy as np
from PIL import Image

__all__ = ['list_images', 'ColorMode', 'color_convert', 'image_to_array', 'array_to_image']


def list_images(directory, file_format=('jpg', 'jpeg', 'bmp', 'png', 'ppm', 'tif', 'tiff')):
    """Lists all pictures in a directory, including all subdirectories.
    
    Args:
        directory: string, absolute path to the directory.
        file_format: tuple of strings or single string, extensions of the pictures.
    Returns:
        a list of image paths.
    """
    file_format = tuple('.%s' % e for e in ((file_format,) if isinstance(file_format, str) else file_format))
    return [os.path.join(root, f) for root, _, files in os.walk(directory) for f in files if f.lower().endswith(file_format)]


class color_mode:
    """image color type."""
    O     = {'mode':'1',      'description':'1-bit pixels, black and white, stored with one pixel per byte'}
    L     = {'mode':'L',      'description':'8-bit pixels, black and white'}
    LA    = {'mode':'LA',     'description':'L with alpha'}
    La    = {'mode':'La',     'description':'L with premultiplied alpha'}
    I     = {'mode':'I',      'description':'32-bit signed integer pixels'}
    F     = {'mode':'F',      'description':'32-bit floating point pixels'}
    P     = {'mode':'P',      'description':'8-bit pixels, mapped to any other mode using a color palette'}
    PA    = {'mode':'PA',     'description':'P with alpha'}
    RGB   = {'mode':'RGB',    'description':'3x8-bit pixels, true color'}
    RGBA  = {'mode':'RGBA',   'description':'4x8-bit pixels, true color with transparency mask'}
    RGBX  = {'mode':'RGBX',   'description':'true color with padding'}
    RGBa  = {'mode':'RGBa',   'description':'true color with premultiplied alpha'}
    CMYK  = {'mode':'CMYK',   'description':'4x8-bit pixels, color separation'}
    YCbCr = {'mode':'YCbCr',  'description':'3x8-bit pixels, color video format'}
    LAB   = {'mode':'LAB',    'description':'3x8-bit pixels, the L*a*b color space'}
    HSV   = {'mode':'HSV',    'description':'3x8-bit pixels, Hue, Saturation, Value color space'}
    I16   = {'mode':'I;16',   'description':'16-bit unsigned integer pixels'}
    I16L  = {'mode':'I;16L',  'description':'16-bit little endian unsigned integer pixels'}
    I16B  = {'mode':'I;16B',  'description':'16-bit big endian unsigned integer pixels'}
    I16N  = {'mode':'I;16N',  'description':'16-bit native endian unsigned integer pixels'}
    BGR15 = {'mode':'BGR;15', 'description':'15-bit reversed true colour'}
    BGR16 = {'mode':'BGR;16', 'description':'16-bit reversed true colour'}
    BGR24 = {'mode':'BGR;24', 'description':'24-bit reversed true colour'}
    BGR32 = {'mode':'BGR;32', 'description':'32-bit reversed true colour'}
    grayscale = {'mode':'grayscale',  'description':'8-bit pixels, black and white'}

ColorMode = color_mode()


def color_convert(image, color_mode=ColorMode.RGB):
    """Transform image color mode.
    
    Args:
        image: PIL Image instance.
        color_mode: Image color mode, more see api "la.image.ColorMode".
    Returns:
        PIL instance.
    Raises:
        ValueError: color_mode error.
    """
    if isinstance(color_mode, dict):
        if color_mode['mode'] == 'grayscale':
            if image.mode not in ('L', 'I;16', 'I'):
                image = image.convert('L')
        elif image.mode != color_mode['mode']:
            image = image.convert(color_mode['mode'])
    else:
        raise ValueError('color_mode error.')
    return image


def image_to_array(image, data_format='HWC', dtype='float32'):
    """Converts a PIL Image instance to a Numpy array.
    
    Args:
        image: PIL instance.
        data_format: array data format, eg.'HWC'.
        dtype: Dtype to use for the returned array.
    Returns:
        A Numpy array.
    Raises:
        ValueError: if invalid `img` or `data_format` is passed.
    """
    transpose = {'H':0, 'W':1, 'C':2}
    x = np.asarray(image, dtype=dtype)
    if len(x.shape) == 2:
        x = np.expand_dims(x, axis=-1)
    if len(x.shape) == 3:
        x = x.transpose(tuple(transpose[i] for i in data_format))
    else:
        raise ValueError('Unsupported image shape: %s' % (x.shape,))
    return x


def array_to_image(x, data_format='HWC'):
    """Converts a 3D Numpy array to a PIL Image instance.
    
    Args:
        x: Input Numpy array.
        data_format: array data format, eg.'HWC'.
    Returns:
        A PIL instance.
    Raises:
        ValueError: if invalid `x` or `data_format` is passed.
    """
    if x.ndim != 3:
        raise ValueError('Expected image array to have rank 3 (single image). '
                         'Got array with shape: %s' % (x.shape,))

    transpose = {i:r for r,i in enumerate(data_format)}
    if data_format != 'HWC':
        x = x.transpose(tuple(transpose[i] for i in 'HWC'))

    if x.shape[2] == 4:
        return Image.fromarray(x.astype('uint8'), 'RGBA')
    elif x.shape[2] == 3:
        return Image.fromarray(x.astype('uint8'), 'RGB')
    elif x.shape[2] == 1:
        if np.max(x) > 255:
            return Image.fromarray(x[:, :, 0].astype('int32'), 'I')
        return Image.fromarray(x[:, :, 0].astype('uint8'), 'L')
    else:
        raise ValueError('Unsupported channel number: %s' % (x.shape[2],))



import json
from collections import defaultdict

import pandas as pd

from linora.utils._config import Config
from linora.feature_column._categorical import *
from linora.feature_column._datetime import *
from linora.feature_column._normalize import *
from linora.feature_column._numerical import *
from linora.feature_column._feature_categorical import FeatureCategorical
from linora.feature_column._feature_datetime import FeatureDatetime
from linora.feature_column._feature_numerical import FeatureNumerical
from linora.feature_column._feature_normalize import FeatureNormalize


__all__ = ['Feature']


class Feature(FeatureCategorical, FeatureDatetime, FeatureNumerical, FeatureNormalize):
    def __init__(self):
        super(Feature, self).__init__()
        self._params = Config()
        self._params.function = {
            'categorical_count':categorical_count, 
            'categorical_crossed':categorical_crossed, 
            'categorical_encoder':categorical_encoder, 
            'categorical_hash':categorical_hash, 
            'categorical_hist':categorical_hist,
            'categorical_regress':categorical_regress,
            'categorical_onehot_binarizer':categorical_onehot_binarizer,
            'categorical_onehot_multiple':categorical_onehot_multiple,
            'categorical_rare':categorical_rare,
            'categorical_regress':categorical_regress,
            'categorical_woe':categorical_woe,
            'datetime_extraction':datetime_extraction,
            'datetime_transform':datetime_transform,
            'numerical_binarizer':numerical_binarizer,
            'numerical_bucketized':numerical_bucketized,
            'numerical_combine':numerical_combine,
            'numerical_cyclical':numerical_cyclical,
            'numerical_math':numerical_math,
            'numerical_padding':numerical_padding, 
            'numerical_outlier':numerical_outlier,
            'numerical_relative':numerical_relative,
            'normalize_max':normalize_max, 
            'normalize_maxabs':normalize_maxabs, 
            'normalize_l1':normalize_l1, 
            'normalize_l2':normalize_l2,
            'normalize_meanminmax':normalize_meanminmax,
            'normalize_minmax':normalize_minmax,
            'normalize_norm':normalize_norm, 
            'normalize_robust':normalize_robust
        }
        self.pipe = {}
        
    def load_pipe(self, json_path):
        """Load pipeline from json format file.
        
        Args:
            json: json format file.
        """
        with open(json, 'r') as f:
            self.pipe = json.load(f)
        return self
        
    def save_pipe(self, json_path):
        """Save pipeline to json format file.
        
        Args:
            json_path: save json file path.
        """
        with open(json_path, 'w') as f:
            json.dump(self.pipe, f)
        return self
        
    def fit(self, df):
        """Fit param value of each variable.
        
        Args:
            df:pd.DataFrame of shape = [n_samples, n_features]
        """
        _ = self._fit_transform(df, fit=True)
        return self
        
    def fit_transform(self, df, keep_columns=None):
        """Fit and transform param value of each variable.
        
        Args:
            df: pd.DataFrame of shape = [n_samples, n_features]
            keep_columns: The original variable that needs to be saved in the final result.
        """
        return self._fit_transform(df, keep_columns=keep_columns, fit=True)
        
    def transform(self, df, keep_columns=None):
        """Transform param value of each variable.
        
        Args:
            df: pd.DataFrame of shape = [n_samples, n_features]
            keep_columns: The original variable that needs to be saved in the final result.
        """
        return self._fit_transform(df, keep_columns=keep_columns, fit=False)
    
    def _fit_transform(self, df, keep_columns=None, fit=True):
        if isinstance(keep_columns, str):
            keep_columns = [keep_columns]
        data = pd.DataFrame() if keep_columns is None else df[keep_columns].copy()
        name_dict = defaultdict(lambda: 0)
        name_drop = [self.pipe[i]['param']['name'] for i in self.pipe if not self.pipe[i]['keep']]
        for r in range(len(self.pipe)):
            config = self.pipe[r].copy()
            t = self._run_function(self._params.function[config['type']], config, df, data, fit)
            if type(t[0])==pd.core.series.Series:
                data[config['param']['name']] = t[0]
                name_dict[config['param']['name']] += 1
            elif type(t[0])==pd.core.frame.DataFrame:
                data = pd.concat([data, t[0]], axis=1)
                for k in t[0].columns:
                    name_dict[k] += 1
            else:
                raise ValueError(f"final result not Series or DataFrame, type error.")
            
            if fit:
                self.pipe[r]['param'] = t[1]['param']
        name_if = [i for i in name_dict if name_dict[i]==1]
        drop = [i for i in name_if if i in name_drop]
        keep = [i for i in data.columns if i not in drop]
        return data[keep]
    
    def _run_function(self, function, config, df, data, fit):
        if config['variable'] in df.columns:
            if 'lable' in config:
                if fit:
                    t = function(df[config['variable']], label=df[config['lable']], **config['param'])
                else:
                    t = function(df[config['variable']], label=df[config['lable']], config=config)
            else:
                if fit:
                    t = function(df[config['variable']], **config['param'])
                else:
                    t = function(df[config['variable']], config=config)
        elif config['variable'] in data.columns:
            if 'lable' in config:
                if fit:
                    t = function(data[config['variable']], label=df[config['lable']], **config['param'])
                else:
                    t = function(data[config['variable']], label=df[config['lable']], config=config)
            else:
                if fit:
                    t = function(data[config['variable']], **config['param'])
                else:
                    t = function(data[config['variable']], config=config)
        else:
            raise ValueError(f"variable `{config['type']}` not exist.")
        return t
        
            
            
            
            
            



from linora.image._image_color import *

class ImageColorAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def enhance_saturation(self, delta, p=None):
        """Adjust image color balance.
        This class can be used to adjust the colour balance of an image, 
        in a manner similar to the controls on a colour TV set. 
        An enhancement factor of 0.0 gives a black and white image. 
        A factor of 1.0 gives the original image.

        Args:
            delta: A floating point value controlling the enhancement. 
                   delta 1.0 always returns a copy of the original image, 
                   lower factors mean less color, 
                   and higher values more. There are no restrictions on this value.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is float multiplier for adjusting color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_saturation(self.image, delta, p)
        return self
    
    def enhance_contrast_linear(self, delta, p=None):
        """Adjust contrast of RGB or grayscale images.
  
        Contrast is adjusted independently for each channel of each image.
        pixel = (x - mean) * delta + mean

        Tips:
            1 means pixel value no change.
            0 means all pixel equal. 
            a suitable interval is (0, 4].
        Args:
            delta: if int, float, a float multiplier for adjusting contrast.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is float multiplier for adjusting contrast.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_contrast_linear(self.image, delta, p)
        return self
    
    def enhance_contrast_gamma(self, gamma=1, gain=1.0, p=None):
        """Perform gamma correction on an image.

        For gamma greater than 1, the histogram will shift towards left and the output image will be darker than the input image. 
        For gamma less than 1, the histogram will shift towards right and the output image will be brighter than the input image.

        pixel = 255*gain *((v/255)**gamma)
        Values in the range gamma=(0.5, 2.0) seem to be sensible.

        Args:
            gamma: float, Non negative real number, gamma larger than 1 make the shadows darker, 
                   while gamma smaller than 1 make dark regions lighter.
                   if list or tuple, randomly picked in the interval `[gamma[0], gamma[1])`.
            gain: float, The constant multiplier.
                  if list or tuple, randomly picked in the interval `[gain[0], gain[1])`.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_contrast_gamma(self.image, gamma, gain, p)
        return self
    
    def enhance_contrast_sigmoid(self, cutoff=0.5, gain=10, p=None):
        """Perform sigmoid correction on an image.

        pixel = 255*1/(1+exp(gain*(cutoff-v/255)))
        Values in the range gain=(5, 20) and cutoff=(0.25, 0.75) seem to be sensible.

        Args:
            image: a PIL instance.
            cutoff: float, Cutoff that shifts the sigmoid function in horizontal direction. 
                    Higher values mean that the switch from dark to light pixels happens later. 
                    if list or tuple, randomly picked in the interval `[cutoff[0], cutoff[1])`.
            gain: float, The constant multiplier.
                  if list or tuple, randomly picked in the interval `[gain[0], gain[1])`.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_contrast_sigmoid(self.image, cutoff, gain, p)
        return self
    
    def enhance_contrast_log(self, gain=1, p=None):
        """Perform log correction on an image.

        pixel = 255*gain*log_2(1+v/255)
        Values in the range gain=[0.6, 1.4] seem to be sensible.

        Args:
            gain: float, The constant multiplier.
                  if list or tuple, randomly picked in the interval `[gain[0], gain[1])`.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_contrast_log(self.image, gain, p)
        return self
    
    def enhance_brightness(self, delta, p=None):
        """Adjust the brightness of RGB or Grayscale images.
    
        Tips:
            delta extreme value in the interval [-1, 1], >1 to white, <-1 to black.
            a suitable interval is [-0.5, 0.5].
            0 means pixel value no change.
        Args:
            delta: if int, float, Amount to add to the pixel values.
                   if list, tuple, randomly picked in the interval `[delta[0], delta[1])` to add to the pixel values.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_brightness(self.image, delta, p)
        return self

    def enhance_sharpness(self, delta, p=None):
        """Adjust image sharpness.
        
        This class can be used to adjust the sharpness of an image. 
        An enhancement factor of 0.0 gives a blurred image, 
        a factor of 1.0 gives the original image, 
        and a factor of 2.0 gives a sharpened image.

        Args:
            delta: A floating point value controlling the enhancement. 
                   delta 1.0 always returns a copy of the original image, 
                   lower factors mean less sharpness, 
                   and higher values more. There are no restrictions on this value.
                   if list, tuple, randomly picked in the interval
                       `[delta[0], delta[1])` , value is float multiplier for adjusting color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_sharpness(self.image, delta, p)
        return self
    
    def enhance_hue(self, delta, p=None):
        """Adjust hue of an image.

        The image hue is adjusted by converting the image to HSV and
        cyclically shifting the intensities in the hue channel (H).
        The image is then converted back to original image mode.

        `delta` is the amount of shift in H channel and must be in the interval `[-1, 1]`.

        Args:
            delta: How much to shift the hue channel. Should be in [-1, 1]. 
                   1 and -1 give complete reversal of hue channel in
                   HSV space in positive and negative direction respectively.
                   0 means no shift. Therefore, both -1 and 1 will give an image
                   with complementary colors while 0 gives the original image.
                   if list or tuple, randomly picked in the interval `[delta[0], delta[1])`.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = enhance_hue(self.image, delta, p)
        return self
    
    def color_invert(self, lower=None, upper=None, wise='pixel', prob=1, p=None):
        """Invert colors of input PIL image.

        Args:
            lower: int or list or tuple, [0, 255], All pixels below this greyscale level are inverted.
                   if list or tuple, randomly picked in the interval `[lower[0], lower[1])`.
            upper: int or list or tuple, [0, 255], All pixels above this greyscale level are inverted.
                   if list or tuple, randomly picked in the interval `[upper[0], upper[1])`.
            wise: 'pixel' or 'channel' or list of channel, method of applying operate.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = color_invert(self.image, lower, upper, wise, prob, p)
        return self
    
    def color_clip(self, lower=None, upper=None, wise='pixel', prob=1, p=None):
        """Clipped colors of input PIL image.

        Args:
            lower: int or list or tuple, [0, 255], All pixels below this greyscale level are clipped.
                   if list or tuple, randomly picked in the interval `[lower[0], lower[1])`.
            upper: int or list or tuple, [0, 255], All pixels above this greyscale level are clipped.
                   if list or tuple, randomly picked in the interval `[upper[0], upper[1])`.
            wise: 'pixel' or 'channel' or list of channel, method of applying operate.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = color_clip(self.image, lower, upper, wise, prob, p)
        return self

    def equalize(self, p=None):
        """
        Equalize the image histogram. This function applies a non-linear
        mapping to the input image, in order to create a uniform
        distribution of grayscale values in the output image.
        
        Args:
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = equalize(self.image, p)
        return self

    def dropout(self, value=0, wise='pixel', prob=0.1, p=None):
        """Drop random channels from images.

        For image data, dropped channels will be filled with value.

        Args:
            value: int or list, dropped channels will be filled with value.
            wise: 'pixel' or 'channel' or list of channel, method of applying operate.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = dropout(self.image, value, wise, prob, p)
        return self


import requests

from linora import gfile
from linora.utils._progbar import Progbar

__all__ = ['get_file']


def assert_dirs(root, root_dir=None, delete=True, make_root_dir=True):
    if root is None:
        root = './'
    assert gfile.isdir(root), '{} should be directory.'.format(root)
    if root_dir is not None:
        assert isinstance(root_dir, str), '{} should be str.'.format(root_dir)
        task_path = gfile.path_join(root, root_dir)
        if gfile.exists(task_path):
            if delete:
                gfile.remove(task_path)
                gfile.makedirs(task_path)
        else:
            if make_root_dir:
                gfile.makedirs(task_path)
        return task_path
    else:
        gfile.makedirs(root)
        return root


def get_file(url, root_file, verbose=1, retries=3, chunk_size=5120):
    """Request url and download to root_file.
    
    Args:
        url: str, request url.
        root_file: str, downloaded and saved file name.
        verbose: Verbosity mode, 0 (silent), 1 (verbose)
        retries: retry counts.
        chunk_size: the number of bytes it should read into memory.
    Return:
        str, downloaded and saved file name.
    """
    for i in range(retries):
        try:
            r = requests.get(url, stream=True)
            content_type = r.headers.get('Content-Length')
            total_size = None if content_type is None else int(content_type.strip())
            p = Progbar(total_size, verbose=verbose)
            down_size = 0
            with open(root_file, 'wb') as f:
                for chunk in r.iter_content(chunk_size):
                    p.add(chunk_size)
                    f.write(chunk)
                    down_size += len(chunk)
            if down_size==total_size:
                break
            raise 'download failed'
        except:
            if i==retries-1:
                raise f'{url} download failed'
    return root_file


import pandas as pd

__all__ = ['uniform', 'quantile', 'probability_categorical']


def uniform(feature, bins):
    """Equal width bin, take a uniform distribution for the sample value range.
    
    Buckets include the right boundary, and exclude the left boundary. 
    Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.], (0., 1.], (1., 2.], and (2., +inf)
    
    Args:
        feature: pd.Series, model feature values.
        bins: int, split bins of feature.
    Returns:
        the list of split threshold of feature.
    """
    t = (feature.max()-feature.min())/bins
    return [t*i for i in range(1, bins)]


def quantile(feature, bins):
    """Equal frequency bin, take a uniform distribution of the sample size.
    
    Buckets include the right boundary, and exclude the left boundary. 
    Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.], (0., 1.], (1., 2.], and (2., +inf).
        
    Args:
        feature: pd.Series, model feature values.
        bins: int, split bins of feature.
    Returns:
        the list of split threshold of feature.
    """
    t = feature.sort_values().values
    w = round(len(t)/bins)
    return [t[w*i-1] for i in range(1, bins)]


def probability_categorical(feature, label):
    """Probability grouping of category variables
    
    Args:
        feature: pd.Series, model feature values.
        label: pd.Series, Target value in supervised learning.
    Returns:
        a dict.{'group': grouping dict, 'data': grouping data, 'distance': grouping distance, 'std': grouping std}
    """
    assert feature.nunique()>2, 'feature category nums must be greater than 2.'
    t = pd.DataFrame({'feature':feature, 'label':label})
    cat = label.unique()
    cat = [(cat[i], cat[i+1]) for i in range(len(cat)-1)]
    prob = label.value_counts(1).to_dict()
    slope = [prob.get(i[0], 0)-prob.get(i[1], 0) for i in cat]
    
    slope_dict = t.feature.value_counts(1).to_dict()
    prob = t.groupby([ 'feature']).label.value_counts(1).to_dict()
    slope_dict = {i:{'category_rate':slope_dict[i], 'slope':[prob.get((i,j[0]), 0)-prob.get((i,j[1]), 0) for j in cat]} for i in slope_dict}
    for i in slope_dict:
        slope_dict[i]['slope_diff'] = sum([abs(slope[j]-slope_dict[i]['slope'][j]) for j in range(len(slope))])
    value1 = sorted([[[i], slope_dict[i]['slope_diff'], slope_dict[i]['category_rate']] for i in slope_dict], key=lambda x:x[1], reverse=1)
    distance = sorted([value1[i][1]-value1[i+1][1] for i in range(len(value1)-1)])
    std = pd.Series([i[1] for i in value1]).std()
    coupe = value1
    dis = distance[0]
    for k in distance:
        value = value1
        while 1:
            for i in range(len(value)-1):
                if value[i][1]-k<value[i+1][1]:
                    value[i+1][0] = value[i][0]+value[i+1][0]
                    value[i+1][1] = value[i][1]*value[i][2]/(value[i][2]+value[i+1][2])+value[i+1][1]*value[i+1][2]/(value[i][2]+value[i+1][2])
                    value[i+1][2] = value[i][2]+value[i+1][2]
                    value.remove(value[i])
                    break
            if i==len(value)-2:
                break
        if pd.Series([i[1] for i in value]).std()>std:
            coupe = value
            std = pd.Series([i[1] for i in value]).std()
            dis = k
    return {'group':{k:i for i,j in enumerate(coupe) for k in j[0]}, 'data':coupe, 
            'distance':dis, 'distance_index':f'{distance.index(dis)+1}/{len(distance)}', 'std':std}


import random
import itertools
from collections import defaultdict

from joblib import Parallel, delayed

__all__ = ['RandomWalker']

class RandomWalker:
    def __init__(self, G):
        """
        :param G: networkx graph instance.
        """
        self.G = G
        
        self._nodes_weight_dict = defaultdict(lambda :defaultdict(list))
        for i in self.G.nodes():
            self._nodes_weight_dict[i]['nodes'] = list(self.G.neighbors(i))
            probs = [self.G[i][j].get('weight', 1.0) for j in self._nodes_weight_dict[i]['nodes']]
            probs_sum = sum(probs)
            self._nodes_weight_dict[i]['weights'] = [i/probs_sum for i in probs]

    def deepwalk_walks(self, walk_num, walk_length, walk_prob=False, filter_lenth=0, workers=1, verbose=0, node_sample=1):
        """Random walk on all nodes
        
        Args:
            walk_num: random walk times, sampling times.
            walk_length: generation sequence length.
            walk_prob: whether each wandering node selection is based on probability.
            filter_lenth: generation sequence length minimum values.
            workers: number of parallel computing processes.
            verbose: whether to enable log output.
            node_sample: (0, 1] is the node random sampling ratio, (1, infinity) is the node random sampling number.
        Return:
            generation sequence list.
        """
        return self._parallel_walks(self._deepwalk_walks, walk_num, walk_length, walk_prob, 
                                    filter_lenth, workers, verbose, node_sample)
        
    def deepwalk_walk(self, walk_length, start_node, walk_prob=False, filter_lenth=0):
        """Random walk on a certain node
        
        Args:
            walk_length: generation sequence length.
            start_node: start wandering node.
            walk_prob: whether each wandering node selection is based on probability.
            filter_lenth: generation sequence length minimum values.
        Return:
            generation sequence list.
        """
        walk = [start_node]
        while len(walk) < walk_length:
            cur = walk[-1]
            cur_nodes = self._nodes_weight_dict[cur]['nodes']
            if len(cur_nodes) > 0:
                if walk_prob:
                    walk.append(random.choices(cur_nodes, self._nodes_weight_dict[cur]['weights'])[0])
                else:
                    walk.append(random.choice(cur_nodes))
            else:
                break
        if len(walk)<=filter_lenth:
            walk = []
        return walk
    
    def _deepwalk_walks(self, nodes, walk_num, walk_length, walk_prob, filter_lenth):
        walks = []
        for _ in range(walk_num):
            random.shuffle(nodes)
            walks += [self.deepwalk_walk(walk_length, v, walk_prob, filter_lenth) for v in nodes]
        return walks
    
    def _parallel_walks(self, method, walk_num, walk_length, walk_prob, filter_lenth, workers, verbose, node_sample):
        nodes = list(self.G.nodes())
        nodes = random.sample(nodes, k=int(len(nodes)*node_sample) if node_sample<=1 else int(node_sample))
        workers_list = [walk_num//workers]*workers + ([walk_num % workers] if walk_num % workers != 0 else [])
        results = Parallel(n_jobs=workers, verbose=verbose)(
            delayed(method)(nodes, num, walk_length, walk_prob, filter_lenth) for num in workers_list)
        walks = itertools.chain(filter(lambda x: len(x) > 0, itertools.chain(*results)))
        return walks



import numpy as np
from PIL import Image, ImageFont, ImageDraw

__all__ = ['grid']


def grid(image, mode, size=None, **kwargs):
    """grid image.

    Args:
        image: a list or tuple of PIL instance.
        mode: grid type.
        size: frist image size.
        rate: if mode in [0,1,2,3], the ratio of the second image, default is 0.25.
        xnums: if mode in [4,5,6,7,8], how many columns to display.
        xspace: if mode in [4,5,6,7,8], width between images in each column.
        yspace: if mode in [4,5,6,7,8], width between images in each row.
        text: if mode in [4,5,6,7,8], each image text, len(text)==len(image)
    Returns:
        a PIL instance.
    """
    if isinstance(image, tuple):
        image = list(image)
    elif not isinstance(image, list):
        image = [image]
    size = image[0].size if size is None else tuple(size)
    for r, i in enumerate(image):
        if size!=i.size:
            image[r] = i.resize(size)
    if mode==0:
        rate = kwargs['rate'] if 'rate' in kwargs else 0.25
        img = image[0].copy()
        size = (int(img.width*rate), int(img.height*rate))
        img.paste(image[-1].resize(size), (img.width-size[0], img.height-size[1]))
        return img
    if mode==1:
        rate = kwargs['rate'] if 'rate' in kwargs else 0.25
        img = image[0].copy()
        size = (int(img.width*rate), int(img.height*rate))
        img.paste(image[-1].resize(size), (img.width-size[0], 0))
        return img
    if mode==2:
        rate = kwargs['rate'] if 'rate' in kwargs else 0.25
        img = image[0].copy()
        size = (int(img.width*rate), int(img.height*rate))
        img.paste(image[-1].resize(size), (0, 0))
        return img
    if mode==3:
        rate = kwargs['rate'] if 'rate' in kwargs else 0.25
        img = image[0].copy()
        size = (int(img.width*rate), int(img.height*rate))
        img.paste(image[-1].resize(size), (0, img.height-size[1]))
        return img
    if mode==4:
        size = image[0].size
        for r, i in enumerate(image):
            if size!=i.size:
                image[r] = i.resize(size)
        xnums = kwargs['xnums'] if 'xnums' in kwargs else np.ceil(np.sqrt(len(image)))
        xspace = kwargs['xspace'] if 'xspace' in kwargs else 5
        yspace = kwargs['yspace'] if 'yspace' in kwargs else 5
        
        shape = [int(xnums), int(np.ceil(len(image)/xnums))]
        img = Image.new(image[0].mode, [shape[0]*(size[0]+xspace)-xspace, shape[1]*(size[1]+yspace)-yspace], (255,255,255))
        for r, i in enumerate(image):
            img.paste(i, (r%shape[0]*(size[0]+xspace), r//shape[0]*(size[1]+yspace)))
        return img
    if mode==5:
        xnums = kwargs['xnums'] if 'xnums' in kwargs else np.ceil(np.sqrt(len(image)))
        text = kwargs['text']
        xspace = kwargs['xspace'] if 'xspace' in kwargs else 5
        yspace = kwargs['yspace'] if 'yspace' in kwargs else 60
        
        shape = [int(xnums), int(np.ceil(len(image)/xnums))]
        img = Image.new(image[0].mode, [shape[0]*(size[0]+xspace)-xspace, shape[1]*(size[1]+yspace)], (255,255,255))
        draw = ImageDraw.Draw(img)
        
        for r, i in enumerate(image):
            img.paste(i, (r%shape[0]*(size[0]+xspace), r//shape[0]*(size[1]+yspace)))
            textsize = draw.textbbox([0,0], text[r])[2:]
            t = min(yspace/2/textsize[1], size[0]/min(textsize[0], textsize[1]*len(text[r])))
            xoffset = int((size[0]-t*min(textsize[0], textsize[1]*len(text[r])))/2)
            font = ImageFont.truetype("Pillow/Tests/fonts/FreeMono.ttf",size=int(t*min(textsize)))
            draw.text((r%shape[0]*(size[0]+xspace)+xoffset, r//shape[0]*(size[1]+yspace)+size[1]+int(yspace/4)), 
                      text[r], fill=0, font=font)
        return img
    if mode==6:
        xnums = kwargs['xnums'] if 'xnums' in kwargs else np.ceil(np.sqrt(len(image)))
        text = kwargs['text']
        xspace = kwargs['xspace'] if 'xspace' in kwargs else 5
        yspace = kwargs['yspace'] if 'yspace' in kwargs else 60
        
        shape = [int(xnums), int(np.ceil(len(image)/xnums))]
        img = Image.new(image[0].mode, [shape[0]*(size[0]+xspace)-xspace, shape[1]*(size[1]+yspace)], (255,255,255))
        draw = ImageDraw.Draw(img)
        
        for r, i in enumerate(image):
            img.paste(i, (r%shape[0]*(size[0]+xspace), r//shape[0]*(size[1]+yspace)+yspace))
            textsize = draw.textbbox([0,0], text[r])[2:]
            t = min(yspace/2/textsize[1], size[0]/min(textsize[0], textsize[1]*len(text[r])))
            xoffset = int((size[0]-t*min(textsize[0], textsize[1]*len(text[r])))/2)
            font = ImageFont.truetype("Pillow/Tests/fonts/FreeMono.ttf",size=int(t*min(textsize)))
            draw.text((r%shape[0]*(size[0]+xspace)+xoffset, r//shape[0]*(size[1]+yspace)+int(yspace/4)), 
                      text[r], fill=0, font=font)
        return img
    if mode==7:
        xnums = kwargs['xnums'] if 'xnums' in kwargs else np.ceil(np.sqrt(len(image)))
        text = kwargs['text']
        xspace = kwargs['xspace'] if 'xspace' in kwargs else 60
        yspace = kwargs['yspace'] if 'yspace' in kwargs else 5
        
        shape = [int(xnums), int(np.ceil(len(image)/xnums))]
        img = Image.new(image[0].mode, [shape[0]*(size[0]+xspace), shape[1]*(size[1]+yspace)-yspace], (255,255,255))
        draw = ImageDraw.Draw(img)
        
        for r, i in enumerate(image):
            img.paste(i, (r%shape[0]*(size[0]+xspace), r//shape[0]*(size[1]+yspace)))
            textsize = draw.textbbox([0,0], text[r])[2:]
            t = min(xspace/2/textsize[1], size[1]/textsize[1]/len(text[r]))
            yoffset = int((size[1]-t*textsize[1]*len(text[r]))/2)
            font = ImageFont.truetype("Pillow/Tests/fonts/FreeMono.ttf",size=int(t*min(textsize)))
            draw.text((r%shape[0]*(size[0]+xspace)+size[0]+int(xspace/4), r//shape[0]*(size[1]+yspace)+yoffset), 
                      text[r], fill=0, font=font, direction='ttb')
        return img
    if mode==8:
        xnums = kwargs['xnums'] if 'xnums' in kwargs else np.ceil(np.sqrt(len(image)))
        text = kwargs['text']
        xspace = kwargs['xspace'] if 'xspace' in kwargs else 60
        yspace = kwargs['yspace'] if 'yspace' in kwargs else 5
        
        shape = [int(xnums), int(np.ceil(len(image)/xnums))]
        img = Image.new(image[0].mode, [size[0]*shape[0]+(shape[0]-1)*xspace, size[1]*shape[1]+shape[1]*yspace], (255,255,255))
        draw = ImageDraw.Draw(img)
        
        for r, i in enumerate(image):
            img.paste(i, (r%shape[0]*(size[0]+xspace)+xspace, r//shape[0]*(size[1]+yspace)))
            textsize = draw.textbbox([0,0], text[r])[2:]
            t = min(xspace/2/textsize[1], size[1]/textsize[1]/len(text[r]))
            yoffset = int((size[1]-t*textsize[1]*len(text[r]))/2)
            font = ImageFont.truetype("Pillow/Tests/fonts/FreeMono.ttf",size=int(t*min(textsize)))
            draw.text((r%shape[0]*(size[0]+xspace)+int(xspace/4), r//shape[0]*(size[1]+yspace)+yoffset), 
                      text[r], fill=0, font=font, direction='ttb')
        return img
    raise ValueError('`mode` value error.')


from linora.image._image_draw import *

class ImageDrawAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def mask(self, size, max_num, random=True, color=None, p=None):
        """Draws a mask.

        Args:
            size: list or tuple, mask size, [height, width]. if int, transform [size, size].
            max_num: int, max mask number.
                     if tuple or list, randomly picked in the interval `[max_num[0], max_num[1])`.
            random: bool, whether the mask position is random.
            color: str or tuple or la.image.RGBMode, rgb color, mask fill color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = mask(self.image, size, max_num, random, color, p)
        return self
    
    def mask_box(self, boxes, fill='inner', color=None, p=None):
        """Fill color inside or outside a rectangular area.

        Args:
            boxes: box axis with same [[x1,y1,x2,y2],...]
            fill: 'inner' or 'outer', color fill position.
            color: str or tuple or la.image.RGBMode, rgb color, box line color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = mask_box(self.image, boxes, fill, color, p)
        return self
    
    def stripe(self, width=4, mode=1, color=None, p=None):
        """vertically and horizontally line apply to image.

        Args:
            width: int, line spacing width.
            mode: 0 is vertically and 1 is horizontally.
            color: int or list, line color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = stripe(self.image, width, mode, color, p)
        return self


import numpy as np

from linora.metrics._utils import _sample_weight

__all__ = ['mean_normal_error', 'mean_absolute_error', 'mean_squared_error',
           'mean_absolute_percentage_error', 'hinge', 'explained_variance_score',
           'median_absolute_error', 'r2_score', 'report_regression',
           'mean_relative_error', 'poisson', 'log_cosh_error', 'max_error',
           'mean_tweedie_deviance', 'mean_poisson_deviance', 'mean_gamma_deviance',
           'mean_pinball_error', 'relative_absolute_error', 'relative_squared_error',
           'relative_normal_error', 'relative_root_mean_squared_error', 'huber_error'
          ]


def mean_normal_error(y_true, y_pred, k, log=False, root=False, sample_weight=None):
    """Mean normal error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        k: int, loss = np.sqrt(loss, 1/k).
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    if log:
        loss = (np.power(np.abs(np.log1p(y_true)-np.log1p(y_pred))*sample_weight, k)).mean()
    else:
        loss = (np.power(np.abs(y_true-y_pred)*sample_weight, k)).mean()
    if root:
        loss = np.power(loss, 1/k)
    return loss


def mean_absolute_error(y_true, y_pred, log=False, sample_weight=None):
    """Mean absolute error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        log: default False, whether to log the variable.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    return mean_normal_error(y_true, y_pred, k=1, log=log, root=False, sample_weight=sample_weight)


def mean_squared_error(y_true, y_pred, log=False, root=False, sample_weight=None):
    """Mean squared error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    return mean_normal_error(y_true, y_pred, k=2, log=log, root=root, sample_weight=sample_weight)


def mean_absolute_percentage_error(y_true, y_pred, sample_weight=None):
    """Mean absolute percentage error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    value = np.abs(y_pred - y_true)
    y_true[np.where(y_true==0)] = 1
    return (value/np.abs(y_true)*sample_weight).mean()


def max_error(y_true, y_pred):
    """The max_error metric calculates the maximum residual error.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
    Returns:
        regression loss values.
    """
    return np.max(np.abs(np.array(y_true) - np.array(y_pred)))


def hinge(y_true, y_pred, k=1, sample_weight=None):
    """hinge regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        k: int, pow() function dim.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return np.power((1-y_true*y_pred).clip(min=0)*sample_weight, k).mean()


def explained_variance_score(y_true, y_pred, sample_weight=None):
    """explained variance regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return 1-((y_true-y_pred)*sample_weight).std()**2/y_true.std()**2


def median_absolute_error(y_true, y_pred, sample_weight=None):
    """Median absolute error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return np.median(np.abs(y_true-y_pred)*sample_weight)


def r2_score(y_true, y_pred, sample_weight=None):
    """r2 regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return 1-np.power((y_true-y_pred)*sample_weight, 2).sum()/np.power((y_true-y_true.mean())*sample_weight, 2).sum()


def report_regression(y_true, y_pred, sample_weight=None, printable=False):
    """regression report
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels.
        sample_weight: list or array of sample weight.
        printable: bool, print report.
    Returns:
        dict, regression report.
    """
    result = {'mean_absolute_error':mean_absolute_error(y_true, y_pred, sample_weight=sample_weight),
              'mean_squared_error':mean_squared_error(y_true, y_pred, sample_weight=sample_weight),
              'mean_absolute_percentage_error':mean_absolute_percentage_error(y_true, y_pred, sample_weight=sample_weight),
              'hinge_loss':hinge(y_true, y_pred, sample_weight=sample_weight),
              'explained_variance_score':explained_variance_score(y_true, y_pred, sample_weight=sample_weight),
              'median_absolute_error':median_absolute_error(y_true, y_pred, sample_weight=sample_weight),
              'r2_score':r2_score(y_true, y_pred, sample_weight=sample_weight)
             }
    if printable:
        print("\nRegression Report")
        print("mean_absolute_error: %.4f" % result['mean_absolute_error'])
        print("mean_squared_error: %.4f" % result['mean_squared_error'])
        print("mean_absolute_percentage_error: %.4f" % result['mean_absolute_percentage_error'])
        print("hinge_loss: %.4f" % result['hinge_loss'])
        print("explained_variance_score: %.4f" % result['explained_variance_score'])
        print("median_absolute_error: %.4f" % result['median_absolute_error'])
        print("r2_score: %.4f" % result['r2_score'])
    return result


def mean_relative_error(y_true, y_pred, normalizer, sample_weight=None):
    """Computes the mean relative error by normalizing with the given values.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        normalizer: The normalizer values with same shape as y_pred.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return (np.abs(y_true-y_pred)/np.array(normalizer)*sample_weight).mean()


def huber_error(y_true, y_pred, delta, sample_weight=None):
    """Computes the huber error between y_true and y_pred.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        delta: int or float, Conditional parameter.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    huber_mse = 0.5 * np.square(y_true - y_pred)
    huber_mae = delta * np.abs(y_true - y_pred) - 0.5 * np.square(delta)
    return np.sum(np.where(np.abs(y_true - y_pred) <= delta, huber_mse, huber_mae)*sample_weight)


def relative_root_mean_squared_error(y_true, y_pred, sample_weight=None):
    """Computes the relative root mean squared error between y_true and y_pred.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return np.sqrt(np.mean(np.square(y_true - y_pred)*sample_weight)/np.sum(np.square(y_pred)))


def relative_absolute_error(y_true, y_pred, sample_weight=None):
    """Computes the relative absolute error between y_true and y_pred.
    
    The relative absolute error is calculated by dividing the total 
    absolute error by the absolute difference between the average value and the actual value.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    return relative_normal_error(y_true, y_pred, k=1, log=False, root=False, sample_weight=sample_weight)


def relative_squared_error(y_true, y_pred, sample_weight=None):
    """Computes the relative squared error between y_true and y_pred.
    
    The relative square error needs to use the mean square error and 
    divide it by the square of the difference between the actual data and the data average.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    return relative_normal_error(y_true, y_pred, k=2, log=False, root=False, sample_weight=sample_weight)


def relative_normal_error(y_true, y_pred, k, log=False, root=False, sample_weight=None):
    """Rrelative normal error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        k: int, loss = np.sqrt(loss, 1/k).
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    if log:
        loss1 = (np.power(np.abs(np.log1p(y_true)-np.log1p(y_pred))*sample_weight, k)).sum()
        loss2 = (np.power(np.abs(np.log1p(y_true)-np.log1p(y_true.mean())), k)).sum()
        loss = loss1/loss2
    else:
        loss = (np.power(np.abs(y_true-y_pred)*sample_weight, k)).sum()/(np.power(np.abs(y_true-y_true.mean()), k)).sum()
    if root:
        loss = np.power(loss, 1/k)
    return loss


def poisson(y_true, y_pred, sample_weight=None):
    """Computes the Poisson loss between y_true and y_pred.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    return np.mean((y_pred - y_true * np.log(y_pred+ 1e-7))*sample_weight)


def log_cosh_error(y_true, y_pred, sample_weight=None):
    """Computes the logarithm of the hyperbolic cosine of the prediction error.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    x = (y_pred-y_true)*sample_weight
    return np.mean(x + np.log(np.exp(-2. * x) + 1.) - np.log(2.))


def mean_tweedie_deviance(y_true, y_pred, p, sample_weight=None):
    """Mean Tweedie deviance regression loss.
    
    when p=0 it is equivalent to mean_squared_error.
    when p=1 it is equivalent to mean_poisson_deviance.
    when p=2 it is equivalent to mean_gamma_deviance.

    p < 0: Extreme stable distribution. Requires: y_pred > 0.
    p = 0 : Normal distribution, output corresponds to mean_squared_error. y_true and y_pred can be any real numbers.
    p = 1 : Poisson distribution. Requires: y_true >= 0 and y_pred > 0.
    1 < p < 2 : Compound Poisson distribution. Requires: y_true >= 0 and y_pred > 0.
    p = 2 : Gamma distribution. Requires: y_true > 0 and y_pred > 0.
    p = 3 : Inverse Gaussian distribution. Requires: y_true > 0 and y_pred > 0.
    otherwise : Positive stable distribution. Requires: y_true > 0 and y_pred > 0.

    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        p: Tweedie power parameter. Either p <= 0 or p >= 1.
           The higher p the less weight is given to extreme deviations between true and predicted targets.
        sample_weight: list or array of sample weight.
    Returns:
        A non-negative floating point value (the best value is 0.0).
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    if p < 0:
        if (y_pred <= 0).any():
            raise ValueError("p < 0: Extreme stable distribution. Requires: y_pred > 0.")
    elif p == 0:
        pass
    elif 0 < p < 1:
        raise ValueError("Tweedie deviance is only defined for p<=0 and p>=1.")
    elif 1 <= p < 2:
        if (y_true < 0).any() or (y_pred <= 0).any():
            raise ValueError("1 < p < 2 : Compound Poisson distribution. Requires: y_true >= 0 and y_pred > 0.")
    else:
        if (y_true <= 0).any() or (y_pred <= 0).any():
            raise ValueError("p>=2, Positive stable distribution. Requires: y_true > 0 and y_pred > 0.")

    if p==0:
        t = np.square(y_true-y_pred)
    elif p==1:
        t = 2*(y_true*np.log(y_true/y_pred)+y_pred-y_true)
    elif p==2:
        t = 2*(np.log(y_pred/y_true)+y_true/y_pred - 1)
    else:
        t = 2 * (
            np.power(np.maximum(y_true, 0), 2 - p) / ((1 - p) * (2 - p))
            - y_true * np.power(y_pred, 1 - p) / (1 - p)
            + np.power(y_pred, 2 - p) / (2 - p))
    sample_weight = _sample_weight(y_true, sample_weight)
    return np.average(t, weights=sample_weight)


def mean_poisson_deviance(y_true, y_pred, sample_weight=None):
    """Mean Poisson deviance regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        A non-negative floating point value (the best value is 0.0).
    """
    return mean_tweedie_deviance(y_true, y_pred, p=1, sample_weight=sample_weight)


def mean_gamma_deviance(y_true, y_pred, sample_weight=None):
    """Mean Gamma deviance regression loss.

    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        A non-negative floating point value (the best value is 0.0).
    """
    return mean_tweedie_deviance(y_true, y_pred, p=2, sample_weight=sample_weight)


def mean_pinball_error(y_true, y_pred, alpha=0.5, sample_weight=None):
    """Pinball loss for quantile regression.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        alpha: this loss is equivalent to Mean absolute error when alpha=0.5, 
               alpha=0.95 is minimized by estimators of the 95th percentile.
        sample_weight: list or array of sample weight.
    Returns:
        A non-negative floating point value (the best value is 0.0).
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    t = alpha*np.maximum(y_true-y_pred, 0)+(1-alpha)*np.maximum(y_pred-y_true, 0)
    return np.average(t, weights=sample_weight)


import datetime
import re
from base64 import b64decode
from datetime import timedelta, timezone, tzinfo
from email.header import decode_header
from quopri import decodestring
from urllib.parse import unquote
from collections import OrderedDict
from collections.abc import Mapping, MutableMapping


class CaseInsensitiveDict(MutableMapping):
    """A case-insensitive ``dict``-like object.
    Inspired by requests.structures
    _store -> 'key.lower()' : ('key','value')
    """

    def __init__(self, data=None, **kwargs):
        self._store = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, _ in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lowerkey, value[1])
            for (lowerkey, value)
            in self._store.items()
        )

    def __eq__(self, other):
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    def copy(self):
        """Shallow copy."""
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))


TYPE_MULTIPART = 'multipart'
TYPE_TEXT_PLAIN = ('text', 'plain')
TYPE_TEXT_HTML = ('text', 'html')
DATE_PATTERN_1 = re.compile(r'(\w+),\s+([0-9]+)\s+(\w+)\s+([0-9]+)\s+([0-9]+):([0-9]+):([0-9]+)\s+(.+)')
DATE_PATTERN_2 = re.compile(r'([0-9]+)\s+([\w]+)\s+([0-9]+)\s+([0-9]+):([0-9]+):([0-9]+)\s+(.+)')
TIMEZONE_PATTERN = re.compile(re.compile(r"([+\-])([0-9])?([0-9])?([0-9])?([0-9])?"))
TIMEZONE_MINUTE_OFFSET = (600, 60, 10, 1)
FILENAME_PATTERN = re.compile(re.compile(r"([^']+)'([^']*)'(.+)"))
MONTH_TO_INT = CaseInsensitiveDict({
    'Jan': 1,
    'Feb': 2,
    'Mar': 3,
    'Apr': 4,
    'May': 5,
    'Jun': 6,
    'Jul': 7,
    'Aug': 8,
    'Sep': 9,
    'Oct': 10,
    'Nov': 11,
    'Dec': 12,
})
HEADER_VALUE_STRIP = '\r\n "'



def recursive_decode(_bytes, encodings):
    """Recursive decode bytes to str."""
    for encoding in encodings:
        try:
            return _bytes.decode(encoding)
        except UnicodeDecodeError:
            pass
    return None


def remove_line_feed_and_whitespace(_value):
    _value = _value.strip(HEADER_VALUE_STRIP)

    while r'\r\n' == _value[:4]:
        _value = _value[4:]
    while r'\r\n' == _value[-4:]:
        _value = _value[:-4]

    return _value


def parse_header_value(bvalue, encodings):
    """Parse mail-specified-header to real value."""
    value = recursive_decode(bvalue, encodings)

    # Decode header without converting charset.
    if value is not None:
        decoded_value = ''
        for _value, _charset in decode_header(value):
            if _charset is not None:
                try:
                    decoded_value += _value.decode(_charset)
                except UnicodeDecodeError:
                    break
            else:
                if isinstance(_value, bytes):
                    decoded_value += _value.decode('utf-8')
                elif isinstance(_value, str):
                    decoded_value += _value
        return decoded_value
    return None


def _fmt_date_tz(tz):
    match_groups = TIMEZONE_PATTERN.match(tz).groups()
    _minute_offset = 0
    if match_groups[0] == '+':
        for i, v in enumerate(match_groups[1:]):
            if v is None:
                continue
            _minute_offset += int(TIMEZONE_MINUTE_OFFSET[i] * int(v))
        return timezone(timedelta(minutes=_minute_offset))
    elif match_groups[0] == '-':
        for i, v in enumerate(match_groups[1:]):
            if v is None:
                continue
            _minute_offset += int(TIMEZONE_MINUTE_OFFSET[i] * int(v))
        return timezone(-timedelta(minutes=_minute_offset))
    return None


def fmt_date(date_as_string):
    """Convert mail header Date to datetime object."""
    match_one = DATE_PATTERN_1.fullmatch(date_as_string)
    match_two = DATE_PATTERN_2.fullmatch(date_as_string)
    if match_one:
        week, day, month_as_string, year, hour, minute, second, time_zone = match_one.groups()
        month = MONTH_TO_INT[month_as_string]
        tz = _fmt_date_tz(time_zone)
        return datetime.datetime(int(year), month, int(day),
                                 int(hour), int(minute), int(second), tzinfo=tz)
    elif match_two:
        day, month_as_string, year, hour, minute, second, time_zone = match_two.groups()
        month = MONTH_TO_INT[month_as_string]
        tz = _fmt_date_tz(time_zone)
        return datetime.datetime(int(year), month, int(day),
                                 int(hour), int(minute), int(second), tzinfo=tz)
    return None


def _get_sub_charset(raw_headers):
    """Hardcode for some invalid mail-encoding."""
    for k, _ in raw_headers:
        if b'X-QQ' in k:
            return ['gbk']
    return []


def parse_headers(lines):
    headers = CaseInsensitiveDict()
    raw_headers = []
    unknown_value_headers = []

    lines_idx = 0
    line = lines[0]
    line_count = len(lines)

    while lines:
        if line in (b'', b'\r\n', b'\n'):
            break
        try:
            bname, bvalue = line.split(b':', 1)
        except:
            raise ValueError('Invalid header:' + str(line))
        bname = bname.strip(b' \t')
        bvalue = bvalue.lstrip()

        # next line
        lines_idx += 1
        if lines_idx <= line_count - 1:
            line = lines[lines_idx]
        else:
            bvalue = bvalue.strip()
            raw_headers.append((bname, bvalue))
            # Parse header.
            name = recursive_decode(bname, ('utf-8',))
            if name is not None:
                value = parse_header_value(bvalue, ('utf-8',))
                if value is not None:
                    headers[name] = value
                else:
                    unknown_value_headers.append((name, bvalue))
            else:
                raise ValueError('Invalid header name {}'.format(str(bname)))
            break

        # consume continuation lines
        continuation = line and line[0] in (32, 9)  # (' ', '\t')

        if continuation:
            bvalue = [bvalue]
            while continuation:
                bvalue.append(line.strip(b' \t'))
                # next line
                lines_idx += 1
                if lines_idx < line_count:
                    line = lines[lines_idx]
                    continuation = line and line[0] in (32, 9)  # (' ', '\t')
                else:
                    line = b''
                    break
            bvalue = b''.join(bvalue)

        bvalue = bvalue.strip()
        raw_headers.append((bname, bvalue))
        # Parse header.
        name = recursive_decode(bname, ('utf-8',))
        if name is not None:
            value = parse_header_value(bvalue, ('utf-8',))
            if value is not None:
                headers[name] = value
            else:
                unknown_value_headers.append((name, bvalue))
        else:
            raise ValueError('Invalid header {}'.format(str(bname)))

    # Parse Content-Type
    try:
        content_type, *extra_pair = headers['content-type'].split(';')
    except Exception as e:
        content_type = 'application/octet-stream'
        extra_pair = []
    main_type, sub_type = content_type.split('/')

    # Remove whitespace and get lower type.
    main_type, sub_type = main_type.replace(' ', '').lower(), sub_type.replace(' ', '').lower()

    # Get extra key values.
    extra_kv = CaseInsensitiveDict()
    for pair in extra_pair:
        if pair:
            try:
                _k, _v = pair.split('=', 1)
                _k = remove_line_feed_and_whitespace(_k)
                _v = remove_line_feed_and_whitespace(_v)
                extra_kv[_k] = _v
            except Exception as e:
                continue

    # Detect charsets
    charsets = []
    main_charset = extra_kv.get('charset')
    if main_charset is None:
        main_charset = 'utf-8'
    else:
        main_charset = main_charset.lower()
    charsets.append(main_charset)

    sub_charset = _get_sub_charset(raw_headers)  # type:list
    for charset in sub_charset:
        if charset not in charsets:
            charsets.append(charset)

    # Re-parses unknown headers
    for name, bvalue in unknown_value_headers:
        value = recursive_decode(bvalue, charsets)
        if value is not None:
            headers[name] = value
    # Parse Date and convert Date to DateTimeObject.
    if headers.get('date'):
        headers['date'] = fmt_date(headers['date'])
    return raw_headers, headers, lines_idx, main_type, sub_type, charsets, extra_kv


def multiple_part_decode(lines, boundary):
    content_text = []
    content_html = []
    attachments = []
    boundary = boundary.encode('ascii')

    # Split to parts.
    parts = []
    part_index = []
    for _idx, line in enumerate(lines):
        if boundary in line:
            part_index.append(_idx)
    if not part_index:
        raise ValueError('Can not find boundary on this mail.boundary{}'.format(boundary.decode("ascii")))
    else:
        _len = len(part_index)
        for idx_idx, idx in enumerate(part_index):
            if idx_idx + 1 <= _len - 1:
                parts.append(lines[idx + 1:part_index[idx_idx + 1]])

    for part in parts:
        parsed_part = parse(part)  # Recursive call
        if parsed_part['content_text']:
            content_text += parsed_part['content_text']
        if parsed_part['content_html']:
            content_html += parsed_part['content_html']
        if parsed_part['attachments']:
            attachments += parsed_part['attachments']

    return content_text, content_html, attachments


def parse_one_part_body(headers, body, main_type, sub_type, transfer_encoding, charsets, extra_kv):
    """Parse non-multiple-part body"""
    transfer_encoding = transfer_encoding.lower()
    content_text = None  # type:None or str
    content_html = None  # type:None or str
    attachment = None  # type:None or tuple

    content_disposition = headers.get('content-disposition')  # type:None or str

    # Is attachment.
    if content_disposition is not None and content_disposition.find('attachment') == 0:
        raw_attachment = _decode_one_part_body(body, transfer_encoding, charsets, _need_decode=False)
        if content_disposition:
            _extra_kv = CaseInsensitiveDict()
            content_disposition_extra_parts = content_disposition.split(';')
            for part in content_disposition_extra_parts:
                if '=' not in part:
                    continue
                try:
                    _k, _v = part.split('=', 1)
                    _k = remove_line_feed_and_whitespace(_k)
                    _v = remove_line_feed_and_whitespace(_v)
                    _extra_kv[_k] = _v
                except Exception as e:
                    continue
            filename = _extra_kv.get('filename')
            if filename is None and _extra_kv.get('filename*'):  # RFC5987 and ignore language tags
                match = FILENAME_PATTERN.fullmatch(_extra_kv.get('filename*'))
                if match:
                    _encoding, _language_tags, _name = match.groups()
                    filename = unquote(_name, _encoding)
        else:
            filename = None
        attachment_name = filename or extra_kv.get('name') or headers.get('subject') or 'Untitled'
        attachment = (attachment_name, raw_attachment)
    # Is text/plain
    elif (main_type, sub_type) == TYPE_TEXT_PLAIN:
        decoded_body = _decode_one_part_body(body, transfer_encoding, charsets)

        if decoded_body:
            content_text = decoded_body
    # Is text/html
    elif (main_type, sub_type) == TYPE_TEXT_HTML:
        decoded_body = _decode_one_part_body(body, transfer_encoding, charsets)
        if decoded_body:
            content_html = decoded_body
    else:  # All other type regard as attachment.
        raw_attachment = _decode_one_part_body(body, transfer_encoding, charsets, _need_decode=False)
        if content_disposition:
            _extra_kv = CaseInsensitiveDict()
            content_disposition_extra_parts = content_disposition.split(';')
            for part in content_disposition_extra_parts:
                if '=' not in part:
                    continue
                try:
                    _k, _v = part.split('=', 1)
                    _k = remove_line_feed_and_whitespace(_k)
                    _v = remove_line_feed_and_whitespace(_v)
                    _extra_kv[_k] = _v
                except Exception as e:
                    continue
            filename = _extra_kv.get('filename')
            if filename is None and _extra_kv.get('filename*'):  # RFC5987 and ignore language tags
                match = FILENAME_PATTERN.fullmatch(_extra_kv.get('filename*'))
                if match:
                    _encoding, _language_tags, _name = match.groups()
                    filename = unquote(_name, _encoding)
        else:
            filename = None
        attachment_name = filename or extra_kv.get('name') or headers.get('subject') or 'Untitled'
        attachment = (attachment_name, raw_attachment)

    return content_text, content_html, attachment


def _decode_one_part_body(lines, transfer_encoding, charsets, _need_decode=True):
    """Decode transfer-encoding then decode raw value to string."""
    if transfer_encoding == 'quoted-printable':
        decoded_bytes = decodestring(b'\r\n'.join(lines))
        if _need_decode:
            return recursive_decode(decoded_bytes, charsets)
        else:
            return b'\r\n'.join(lines)
    elif transfer_encoding == 'base64':
        decoded_bytes = b64decode(b''.join(lines))
        if _need_decode:
            return recursive_decode(decoded_bytes, charsets)
        else:
            return decoded_bytes
    elif transfer_encoding in ('binary', '8bit', '7bit'):
        if _need_decode:
            return recursive_decode(b'\r\n'.join(lines), charsets)
        else:
            return b'\r\n'.join(lines)
    else:
        raise ValueError('Invalid transfer-encoding {}'.format(transfer_encoding))


def parse(lines):
    """Decode a multiple-part or Non-multiple-part mail to ParsedMail(as CaseInsensitiveDict)."""
    content_text = []
    content_html = []
    attachments = []

    raw_headers, headers, eof_idx, main_type, sub_type, charsets, extra_kv = parse_headers(lines)

    body = lines[eof_idx + 1:]

    if main_type == TYPE_MULTIPART:  # Include recursive call
        boundary = extra_kv.get('boundary')
        if boundary is None:
            raise ValueError('Can not find boundary in multiple-part mail.')

        _content_text, _content_html, _attachment = multiple_part_decode(body, boundary)

        if _content_text:
            content_text += _content_text
        if _content_html:
            content_html += _content_html
        if _attachment:
            attachments += _attachment

    else:  # Recursive exit
        transfer_encoding = headers.get('content-transfer-encoding')

        if transfer_encoding is not None:
            transfer_encoding = transfer_encoding.lower()
        else:
            # Default transfer_encoding.
            transfer_encoding = '8bit'

        _content_text, _content_html, _attachment = parse_one_part_body(headers, body, main_type, sub_type,
                                                                        transfer_encoding, charsets, extra_kv)
        if _content_text:
            content_text.append(_content_text)
        if _content_html:
            content_html.append(_content_html)
        if _attachment:
            attachments.append(_attachment)

    mail = dict()
    mail['content_text'] = content_text
    mail['content_html'] = content_html
    
#     attachments = [i for i in attachments if i[0]!='Untitled'] 
    mail['attachments'] = attachments
    
#     mail['headers'] = headers
#     mail['raw_headers'] = raw_headers
#     mail['charsets'] = charsets

    mail['subject'] = headers.get('Subject')
    mail['date'] = headers.get('Date')
    mail['from'] = headers.get('From')
    mail['to'] = headers.get('To')
    mail['cc'] = headers.get('Cc')
    return mail


def parse_mail(lines, which):
    """A wrapper for parse mail."""
    mail = parse(lines)
    mail['id'] = which
    if mail['to'] is not None:
        mail['to'] = [i.strip() for i in str(mail['to']).split(',')]
    if mail['cc'] is not None:
        mail['cc'] = [i.strip()[1:-1] for i in str(mail['cc']).split(',')]
#     parsed_mail['raw'] = lines
    return mail